/*      file     : TrafficData/MicroSimCA.C
        project  : PAMINA III
        author   : Marcus Rickert
        comments : multilane CA rules
        RCS      : $Id: MicroSimCA.C,v 1.10 2005/03/25 12:28:51 mr Exp $

        Copyright (C) 2005 Marcus Rickert

        http://www.the-rickerts.de/mr/PAMINA.html
*/

// The update rules in this file are designed to be part of a two sub step update


// ***DEFINES*************************************************************************

#define SITE_ROUTE_CHANGE_LEFT_PERMITTED(CLane, Site) \
           ((((Site) & SITE_MOVE_RIGHT) == 0) OR (CLane) > SITE_GET_BND_LANE(Site))
#define SITE_ROUTE_CHANGE_RIGHT_PERMITTED(CLane, Site) \
           ((((Site) & SITE_MOVE_LEFT) == 0) OR (CLane) < SITE_GET_BND_LANE(Site))
#define SITE_ROUTE_CHANGE_LEFT_DESIRED(CLane, Site) \
           ((((Site) & SITE_MOVE_LEFT) != 0) AND (CLane) > SITE_GET_BND_LANE(Site))
#define SITE_ROUTE_CHANGE_RIGHT_DESIRED(CLane, Site) \
           ((((Site) & SITE_MOVE_RIGHT) != 0) AND (CLane) < SITE_GET_BND_LANE(Site))

// ***METHODS************************************************************************* 

// 'LaneChange' does a symmetric exchange of vehicles. The input data is in
// Sites[FirstSite-2*GRID_BND_WIDTH : LastSite+2*GRID_BND_WIDTH[. The valid output
// data is in Scratch[FirstSite-GRID_BND_WIDTH : LastSite+GRID_BND_WIDTH[.

void
TMultilaneGrid::LaneChange(void)

{
    long  CLane;
    long  Pred[MAX_LANES];
    long  Current;
    long  Lookback = 0;
    long  MaxLookback = 0;
    long  CV;
    TSite Site;
    long  ChangeInfo;

    LaneChangeArray.Resize(0);

    for (CLane = 0; CLane < Config.NrOfLanes; CLane++)
    {
        // Look for a predeccessor in [LastSite : LastSite + ChangeViewRange[
        // After the loop Pred[CLane] points to the predecessor if there is one. If there 
        // isn't it points to LastSite+ChangeViewRange.

        Pred[CLane] = LastActiveSite;
        while (Pred[CLane] < LastActiveSite + GetChangeViewRange() AND 
               !SITE_IS_OCCUPIED(SAccess(CLane, Pred[CLane])))
            Pred[CLane]++;
    }

    // Update (downward!) [FirstSite : LastSite[
    
    for (Current = LastActiveSite - 1; Current >= FirstActiveSite; Current--)
    {
        for (CLane = Config.NrOfLanes - 1; CLane >= 0; CLane--)
        {
            if (SITE_IS_OCCUPIED(SAccess(CLane, Current)))
            {
                if (SITE_IS_VEHICLE(Site = SAccess(CLane, Current)))
                {                   
                    // goto NoChange;
                
                    CV = SITE_GET_VELOCITY(Site);
                    
                    // Check changing left 
                    
                    if (CLane < Config.NrOfLanes - 1 AND
                        SITE_CHANGE_LEFT_PERMITTED(Site) AND
#ifdef CA_VEHICLE_IDS
                        SITE_ROUTE_CHANGE_LEFT_PERMITTED(CLane, Site) AND
#endif
                        ((
#ifdef LATOUR_MODEL_3
                          PredLGap(CLane, Current) < CV AND     // T1
                          PredLGap(CLane + 1, Current) >= CV    // T2
#else
                          PredLGap(CLane, Current) < CV + 1 AND // T1
                          PredLGap(CLane + 1, Current) > CV     // T2
#endif
                          )
#ifdef CA_VEHICLE_IDS
                         OR
                         SITE_ROUTE_CHANGE_LEFT_DESIRED(CLane, Site)
#endif
                         )
                        )
                        
                    {
                        // Check the site beside us
                        
                        if (SITE_IS_OCCUPIED(SAccess(CLane + 1, Current)))
                            goto CheckRight;
                        
                        // Now we have to check the road looking over our left shoulder.
                        // We check SITE_V_MAX sites. If we detect a vehicle we won't
                        // change lanes. 

                        if (SITE_ROUTE_CHANGE_LEFT_DESIRED(CLane, Site))
                            MaxLookback = min(Config.RouteLookBack[CHANGE_LEFT], 
                                              SITE_GET_V_MAX(Site));
                        else
                            MaxLookback = Config.LookBack[CHANGE_LEFT];
                        Lookback = 0; 
                        while (++Lookback <= MaxLookback)   // T3
                            if (SITE_IS_VEHICLE(SAccess(CLane + 1, 
                                                        Current - Lookback)))
                                goto CheckRight;
                        if (SITE_ROUTE_CHANGE_LEFT_DESIRED(CLane, Site) OR
                            Random->IsLower(ChangeProbLimit[CHANGE_LEFT])    // T4
                            )
                        {
                            ChangeInfo = LC_COMPOSE(Current, CLane, CLane + 1);
                            LaneChangeArray.Add(ChangeInfo);
#ifdef CA_MONITOR_PING_PONG
                            if (SITE_IS_PREV_TS_LANE_CHANGE(Site))
                                if (CV >= FAST_PING_PONG_V_MIN)
                                    FastPingPongLaneChanges[CLane][CHANGE_LEFT]++;
                                else
                                    SlowPingPongLaneChanges[CLane][CHANGE_LEFT]++;
#endif
#ifdef CA_MONITOR_LANE_CHANGES                      
                            LaneChanges[CLane][CHANGE_LEFT]++;
#endif
                            // Mark the current vehicle as predecessor for the next vehicle 
                            
                            Pred[CLane] = Current;
                            goto Finished;
                        }
                    }
                    
                CheckRight:  // I know it's very bad style. But this is REALLY a label!
                    
                    
                    if (CLane > 0 AND
                        SITE_CHANGE_RIGHT_PERMITTED(Site) AND
#ifdef CA_VEHICLE_IDS
                        SITE_ROUTE_CHANGE_RIGHT_PERMITTED(CLane, Site) AND
#endif
                        (((IsAsymetric() OR                       // exclude T1 for L->R
                           PredLGap(CLane, Current) < CV + 1) AND // T1
#ifdef LATOUR_MODEL_3                     
                          PredLGap(CLane - 1, Current) >= CV      // T2
#else
                          PredLGap(CLane - 1, Current) > CV       // T2
#endif
                          )      
#ifdef CA_VEHICLE_IDS
                         OR
                         SITE_ROUTE_CHANGE_RIGHT_DESIRED(CLane, Site)
#endif
                         ))
                    {
                        // Check the site beside us
                        
                        if (SITE_IS_OCCUPIED(SAccess(CLane - 1, Current)))
                            goto Finished;
                        
                        // Now we have to check the road looking over our right shoulder.
                        // We check SITE_V_MAX sites. If we detect a vehicle we won't
                        // change lanes. 
                        
                        if (SITE_ROUTE_CHANGE_RIGHT_DESIRED(CLane, Site))
                            MaxLookback = min(Config.RouteLookBack[CHANGE_RIGHT], 
                                              SITE_GET_V_MAX(Site));
                        else
                            MaxLookback = Config.LookBack[CHANGE_RIGHT];
                        Lookback = 0; 
                        while (++Lookback <= MaxLookback)        // T3
                            if (SITE_IS_VEHICLE(SAccess(CLane - 1, Current - Lookback)))
                                goto NoChange;
                        if (SITE_ROUTE_CHANGE_RIGHT_DESIRED(CLane, Site) OR
                            Random->IsLower(ChangeProbLimit[CHANGE_RIGHT])   // T4
                            )
                        {
                            ChangeInfo = LC_COMPOSE(Current, CLane, CLane - 1);
                            LaneChangeArray.Add(ChangeInfo);
#ifdef CA_MONITOR_PING_PONG
                            if (SITE_IS_PREV_TS_LANE_CHANGE(Site))
                                if (SITE_GET_VELOCITY(Site) >= FAST_PING_PONG_V_MIN)
                                    FastPingPongLaneChanges[CLane][CHANGE_RIGHT]++;
                                else
                                    SlowPingPongLaneChanges[CLane][CHANGE_RIGHT]++;
#endif
#ifdef CA_MONITOR_LANE_CHANGES                     
                            LaneChanges[CLane][CHANGE_RIGHT]++;
#endif
                            // The next statement is pretty ugly. It breaks the symmetry between 
                            // the directions left->right and right->left for more than two lanes, 
                            // although the update rules are symmetric in principal. 
                            // The reason is that
                            // a center lane (only occuring if there are more than two lanes) might
                            // receive a vehicle from both directions. Here we check whether there
                            // is a lane two lanes to the right and temporarily prohibit the vehicle
                            // occupying that site to change left in the current time step.
                        
                            if (CLane >= 2)
                                SAccess(CLane - 2, Current) |= SITE_TMP_NO_CHANGE_LEFT;
                            goto Finished;
                        }
                    }

                NoChange: ;

#ifdef CA_MONITOR_PING_PONG
                    SAccess(CLane, Current) &= ~SITE_PREV_TS_LANE_CHANGE;
#endif              
                } // if (SITE_IS_VEHICLE)               
            
            Finished:
                
                Pred[CLane] = Current;
            } // if (SITE_IS_OCCUPIED)
        } // for all lanes
    } // for all sites
}

// =================================================================================== 

void
TMultilaneGrid::Move()

{
    long   CLane;
    long   Current;
    TSite *PCurrent = 0;
    long   Pred = 0;
    long   CV;
    TSite  Site;
    long   Vmax;

    for (CLane = 0; CLane < Config.NrOfLanes; CLane++)
    {
        // Look for a predeccessor in [LastSite : LastSite + MoveViewRange[
        // After the loop Pred points to the predecessor if there is one. If there isn't
        // it points to LastSite+MoveViewRange.
        
        Pred = LastActiveSite;
        while (Pred < LastActiveSite + GetMoveViewRange() AND 
               !SITE_IS_OCCUPIED(SAccess(CLane, Pred)))
            Pred++;
        
        // Update (downward!) [FirstSite : LastSite[

        PCurrent = & SAccess(CLane, LastActiveSite - 1);
        for (Current = LastActiveSite - 1; Current >= FirstActiveSite; Current--)
        {

            if (SITE_IS_OCCUPIED(*PCurrent))
            {
                if (SITE_IS_VEHICLE(*PCurrent))
                {
                    CV = SITE_GET_VELOCITY(Site = *PCurrent);
                    Vmax = SITE_GET_V_MAX(Site);
                    
                    // First rule of Kai's CA
                    
                    if (CV < Vmax)  // vehicle's maximum velocity
                    {
                        if (CV < SpeedLimit) // grid's maximum velocity
                            CV++;
                    }
                    else
                        CV = Vmax; // make sure Vmax is used if it was changed
        
                    // Second rule
                    
                    if (CV > PredGap(Current))
                        CV = PredGap(Current);
                    
                    // Third rule
        
#ifndef CA_DETERMINISTIC
                    if (CV AND RandomDeceleration(Site))
                        CV--;
#endif
                    
                    // Move the vehicle to the new position in the Sites Array. 
        
                    UpdateSite(CLane, Current, CV);

                }
                // Mark the current vehicle as predecessor for the next vehicle down the road
                
                Pred = Current;
            }
            PCurrent--;
        }
    }
}


