/*      file     : TrafficData/TGridStat.C
        project  : PAMINA III
        author   : Marcus Rickert
        RCS      : $Id: TGridStat.C,v 1.9 2005/03/25 12:28:51 mr Exp $

        Copyright (C) 2005 Marcus Rickert

        http://www.the-rickerts.de/mr/PAMINA.html
*/

/* ***INCLUDES************************************************************************ */

#include <math.h>

#include <TLocation.h>
#include <TBaseStream.h>
#include <TTimeStep.h>
#include <TEdge.h>
#include <TEdgeInfo.h>

#include "Transims.h"
#include "TMultilaneGrid.h"
#include "TGridStat.h"

// ***GLOBAL VARIABLES****************************************************************  

TClassInfo *TGridStat::ClassInfo = 0;
bool        TGridStat::NewTransimsFeedback = FALSE;

/* ***METHODS************************************************************************* */

#ifdef NO_INLINE
#  include <TGridStat.i.C>
#endif

/* =================================================================================== */
#define ParentClass TStatObject
/* =================================================================================== */

TGridStat::TGridStat(TParallelID theParallelID) :

ParentClass(ClassInfo, theParallelID)

{
    Reset();
}

/* =================================================================================== */

TGridStat::~TGridStat(void)

{
}

/* =================================================================================== */

/* The method 'Add' must be overwritten to ensure that two statistic objects (here
   the current object and the object passed as a parameter) can
   be merged into one single object (which is the current object) ususally by
   adding or multiplying local variables. But of course any calculation can be
   performed that is both commutative and associative. */

bool
TGridStat::Accumulate(const TStatObject *StatObject)

{
    const TGridStat *Stat = (const TGridStat *) StatObject;

    SampleCount += Stat->SampleCount;
    Length += Stat->Length;
    Sites += Stat->Sites;
    Count += Stat->Count;
    SumV  += Stat->SumV;
    SumV2 += Stat->SumV2;
    AverageDelay += Stat->AverageDelay;
    return ParentClass::Accumulate(StatObject);
}


/* =================================================================================== */

bool
TGridStat::Collect(const TMultilaneGrid *Grid, long theAverageDelay)

{
    long CLane, CSite;

    SampleCount++;
    Length = Grid->GetNrOfLocalSites();
    FreeSpeed = SITE_LENGTH * (1.0 * Grid->Config.SpeedLimit - Grid->Config.DecProb);
    for (CLane = 0; CLane < Grid->GetNrOfLanes(); CLane++)
    {
        Sites += Grid->GetNrOfLocalSites();
        for (CSite = Grid->GetFirstSite(); 
             CSite < Grid->GetLastSite() + Grid->GetBndWidth(); CSite++)
            if (Grid->VAccess(CLane, CSite) != 0)
            {
                int V;

                Count++;
                SumV += (V = SITE_GET_VELOCITY(Grid->SAccess(CLane, CSite)));
                SumV2 += V*V;
            }
    }
    AverageDelay += theAverageDelay;
    return TRUE;
}

/* =================================================================================== */

bool
TGridStat::Decode(TBaseStream *aStream, TTransferInfo *TransferInfo)

{
    return (ParentClass::Decode(aStream, TransferInfo) AND
            aStream->Decode(Length, GRIDSTAT_LONG) AND
            aStream->Decode(FreeSpeed, GRIDSTAT_FLOAT));
}

/* =================================================================================== */

bool
TGridStat::Encode(TBaseStream *aStream, TTransferInfo *TransferInfo) const

{
    return (ParentClass::Encode(aStream, TransferInfo) AND
            aStream->Encode(Length, GRIDSTAT_LONG) AND
            aStream->Encode(FreeSpeed, GRIDSTAT_FLOAT));
}

/* =================================================================================== */

TString
TGridStat::GetClassName(void) const

{
    return "TGridStat";
}

/* =================================================================================== */

bool
TGridStat::GetConsistentTravelTime(long &VehicleCount, double &TravelTime) const

{
    double VMean;
    double Meters;

    if (!GetConsistentVMean(VMean))
    {
        Error("GetConsistentTT: cannot get consistent VMean");
        return FALSE;
    }
    Meters = GetLengthInMeters();
    TravelTime = Meters / VMean;
    if (GetNrOfVehicles() >= TRANSIMS_LINK_STAT_MIN_COUNT)
        VehicleCount = GetNrOfVehicles();
    else
        VehicleCount = 1;
    return TRUE;
}

/* =================================================================================== */

bool
TGridStat::GetConsistentVMean(double &VMean) const

{
    if (GetSumV() > 0)
    {
        VMean = KMH2MPS(CA_VELOCITY_TO_REAL_VELOCITY(GetVMean()));
        if (VMean < TRANSIMS_LINK_STAT_MIN_SPEED)
            VMean = TRANSIMS_LINK_STAT_MIN_SPEED;
    }
    else
        if (GetDensity() > TRANSIMS_LINK_STAT_GRIDLOCK_DENSITY)
            VMean = TRANSIMS_LINK_STAT_MIN_SPEED;
        else
        {
            TEdge     *Edge;
            TEdgeInfo *EdgeInfo;
            
            if ((Edge = (TEdge *) TEdge::ClassInfo->GetObjectPtr(GetParallelID())) == 0)
            {
                Error("WriteTransimsLinkStat: cannot retrieve edge for %ld", GetParallelID());
                return FALSE;
            }
            if ((EdgeInfo = (TEdgeInfo *) Edge->GetInfo(TEdgeInfo::ClassInfo->GetInfoID(), 
                                                        TRUE, TRUE)) == 0)
            {
                Error("WriteTransimsLinkStat: cannot retrieve edge info for %s", 
                      (char *) Edge->GetObjectName());
                return FALSE;
            }
            VMean = EdgeInfo->FreeSpeed;
        }
    return TRUE;
}

/* =================================================================================== */

double
TGridStat::GetVDelta(void) const

{
    if (GetNrOfVehicles())
        return sqrt((double) SumV2 * (double) GetNrOfVehicles() -
                    (double) SumV * (double) SumV) / (double) GetNrOfVehicles();
    else
        return INVALID_MEAN;
}

/* =================================================================================== */

void
TGridStat::Reset(void)

{
    ParentClass::Reset();
    SampleCount = 0;
    Length = 0;
    Sites = 0;
    Count = 0;
    SumV = 0;
    SumV2 = 0;
    AverageDelay = 0;
}

/* =================================================================================== */

bool
TGridStat::WriteTransimsLinkStat(TBaseStream *Stream)

{
    TEdge     *Edge;
    long       EffectiveCount;
    double     TravelTime;

    if ((Edge = (TEdge *) TEdge::ClassInfo->GetObjectPtr(GetParallelID())) == 0)
    {
        Error("WriteTransimsLinkStat: cannot retrieve edge for %ld", GetParallelID());
        return FALSE;
    }
    if (NewTransimsFeedback)
    {
        if (!GetConsistentTravelTime(EffectiveCount, TravelTime))
        {
            Error("WriteTransimsLinkStat: cannot get TT for %s", 
                  (char *) Edge->GetObjectName());      
            return FALSE;
        }
    }
    else
    {
        double EffectiveLength;
        double Meters;
        double AverageSpeed;
        double EffectiveSumV;

        if ((EffectiveLength = 
             Edge->GetLength(TAdjacenceInfo::ClassInfo->GetInfoID())) == (double) INVALID)
        {
            Error("cannot retrieve length for %s", (char *) Edge->GetObjectName());
            return FALSE;
        }
        Meters = LengthInMeter(EffectiveLength);
        if (GetNrOfVehicles() >= TRANSIMS_LINK_STAT_MIN_COUNT)
        {
            EffectiveCount = GetNrOfVehicles();
            EffectiveSumV = KMH2MPS(CA_VELOCITY_TO_REAL_VELOCITY(GetSumV()));
        }
        else
        {
            TEdgeInfo *EdgeInfo;
            
            if ((EdgeInfo = (TEdgeInfo *) 
                 Edge->GetInfo(TEdgeInfo::ClassInfo->GetInfoID(), TRUE, TRUE)) == 0)
            {
                Error("WriteTransimsLinkStat: cannot retrieve edge info for %s", 
                      (char *) Edge->GetObjectName());
                return FALSE;
            }
            EffectiveCount = 1;
            EffectiveSumV = EdgeInfo->FreeSpeed;
        }
        AverageSpeed = EffectiveSumV / EffectiveCount;
        if (AverageSpeed < TRANSIMS_LINK_STAT_MIN_SPEED)
            AverageSpeed = TRANSIMS_LINK_STAT_MIN_SPEED;
        TravelTime = Meters / AverageSpeed;
    }
    if (AverageDelay > 0)
    {
        long RealAverageDelay;
        
        RealAverageDelay = (long) ((double) AverageDelay / SampleCount);
        TravelTime += RealAverageDelay;
    }
    Stream->Write("%.2lf %ld %ld %.2lf %ld %lu\n",
                  (double) (EffectiveCount * TravelTime),              // travel time on link in meters/sec
                  (long)   TimeStep,                                   // sampling time step
                  (long)   Edge->GetFromNodeID(), 
                  (double) (EffectiveCount * TravelTime * TravelTime), // faked 'travel time'^2 (no variance)
                  (long)   EffectiveCount,                             // number of vehicles on link
                  (long)   PaminaEdge2TransimsLink(GetParallelID()));
    return TRUE;
}

/* =================================================================================== */
#undef ParentClass
