/*      file     : TrafficData/TRouter.C
        project  : PAMINA III
        author   : Marcus Rickert
        RCS      : $Id: TRouter.C,v 1.13 2005/03/29 17:12:48 mr Exp $

        Copyright (C) 2005 Marcus Rickert

        http://www.the-rickerts.de/mr/PAMINA.html
*/

// ***INCLUDES************************************************************************ 

#include <math.h>
#include <stdlib.h>

#include <TParser.h>

#include <TEnvironment.h>
#include <TParallel.h>
#include <TCommunication.h>
#include <TBaseMessage.h>
#include <TTransferInfo.h>
#include <TExtendedCPNInfo.h>
#include <TNetwork.h>
#include <TObjectManager.h>
#include <TStatManager.h>
#include <TSamplingDescription.h>
#include <TAggregationInfo.h>
#include <TAggregationResult.h>

#include <TrafficDataTypes.h>
#include <TrafficDataConfig.h>
#include <TConfiguration.h>
#include <TNode.h>
#include <TEdge.h>
#include <TVehicle.h>
#include <TAccessory.h>
#include <TSimpleIntersection.h>
#include <TMapStat.h>
#include <TTrafficTimers.h>

#include "Transims.h"
#include "TShortestPath.h"
#include "TRoutingStat.h"
#include "TGridStat.h"
#include "TGridStatLookup.h"
#include "TRouteplan.h"
#include "TMultilaneGrid.h"
#include "TRouter.h"

// ***DEFINES************************************************************************* 

#define REROUTE_ERROR_BUILD_TREE                -1
#define REROUTE_ERROR_GET_PATH                  -2
#define REROUTE_TYPE_TOO_CLOSE_TO_DEST           0
#define REROUTE_TYPE_TOO_SHORT                   1
#define REROUTE_TYPE_REROUTE                     2
#define REROUTE_TYPE_DETOUR                      3
#define REROUTE_RESULT_KEEP_ROUTE                1
#define REROUTE_RESULT_SAME_ROUTE                2
#define REROUTE_RESULT_NEW_ROUTE                 3

#define CONVERSION_LOGFILE_EXTENSION  ".log"
#define CONVERSION_STATFILE_EXTENSION ".cvs"
#define ROUTING_STATFILE_EXTENSION    ".rts"
#define COMPLETED_ROUTES_EXTENSION    ".crt"
#define ROUTES_BY_ORIGIN_EXTENSION    ".rbo"
#define REROUTING_LOGFILE_EXTENSION   ".rrl"
#define COMMENT_CHAR                  '#'
#define ROUTE_BUFFER_SIZE             16384
#define OFFSET_FILENAME               "/tmp/RouteConverter_offset_%D.tmp"
#define ROUTEPLAN_FILENAME            "/tmp/RouteConverter_route_%D.tmp"

#define REROUTE_MIN_REMAINING_LENGTH             2

#define ROUTER_DEFAULT_BINARYROUTEPLAN           0L

#define ROUTER_DEFAULT_GRIDSTATCOLLECTINTERVAL   TRANSIMS_LINK_STAT_INTERVAL
#define ROUTER_DEFAULT_GRIDSTATSAMPLEINTERVAL    60L
#define ROUTER_DEFAULT_MAXGRIDSTATENTRIES        32L
#define ROUTER_DEFAULT_LOCALGRIDSTATOVERLAP      0.0 // 1 [km]

#define ROUTER_DEFAULT_ROUTINGSTATINTERVAL       20L
#define ROUTER_DEFAULT_ROUTINGSTATDUMP           0L
#define ROUTER_DEFAULT_ROUTEPLANEXTENSION        ""

#define ROUTER_DEFAULT_COLLECTROUTESBYORIGIN     0L
#define ROUTER_DEFAULT_ROUTEORIGINSECTORS        4L

#define ROUTER_DEFAULT_TRAVELTIMEESTIMATEUPDATE  0L

// ***GLOBAL VARIABLES****************************************************************  

TRouter *Router = 0;

// ***CONSTANTS***********************************************************************

const char *
RouteErrors[ROUTE_ERRORS] =

{
    "OK",
    "direction undeclared (edge %ld)",
    "too short",
    "no common node (edges %ld -> %ld)",
    "missing accessory/ies (edge %ld)"
};

// ***METHODS************************************************************************* 

int
CompareRouteInfos(TRouteInfo *Info1, TRouteInfo *Info2)

{
    if (Info1->DepartureTime < Info2->DepartureTime)
        return -1;
    else
        if (Info1->DepartureTime > Info2->DepartureTime)
            return 1;
        else
            return 0;
}

// =================================================================================== 
#define ParentClass TBaseObject
// =================================================================================== 

TRouter::TRouter(void)

{
    GridAggregationInfo = LocalGridAggregationInfo = 0;
    LocalSP = 0;
    RouteplanParser = 0;
    RouteplanOutput = 0;
    ConversionStatOutput = 0;
    ConversionLogOutput = 0;    
    RouteplanTempOutput = 0;
    CompletedRoutesOutput = 0;
    RoutesByOriginOutput = 0;
    TransimsLinkStatOutput = 0;
    OffsetFile = 0;    
    RoutingStatOutput = 0;
    LatestRoutingStat = 0;
    ReroutingLog = 0;
    SkipRouteplan = 0;
    SkipVehicle = 0;
    LocalGridStatValid = FALSE;
    Reset();
}

// =================================================================================== 

TRouter::~TRouter(void)

{
    DeleteObject(LocalSP);
    DeleteObject(LatestRoutingStat);
    DeleteObject(SkipRouteplan);
    DeleteObject(SkipVehicle);
    CloseFiles();
    CloseRouteplanInput();
    Router = 0;
}

// ===================================================================================

bool
TRouter::Activate(void)

{
    if (Router)
    {
        Error("another instance of TRouter is already active");
        return FALSE;
    }
    Router = this;
    Configuration->SetDefault(TConfigEntry(ROUTER_GROUP, ROUTER_NAME_ROUTINGSTATINTERVAL,
                                           ROUTER_DEFAULT_ROUTINGSTATINTERVAL));
    Configuration->SetDefault(TConfigEntry(ROUTER_GROUP, ROUTER_NAME_ROUTEPLANEXTENSION,
                                           ROUTER_DEFAULT_ROUTEPLANEXTENSION));
    Configuration->SetDefault(TConfigEntry(ROUTER_GROUP, ROUTER_NAME_ROUTINGSTATDUMP,
                                           ROUTER_DEFAULT_ROUTINGSTATDUMP));
    Configuration->SetDefault(TConfigEntry(ROUTER_GROUP, ROUTER_NAME_GRIDSTATSAMPLEINTERVAL,
                                           ROUTER_DEFAULT_GRIDSTATSAMPLEINTERVAL));
    Configuration->SetDefault(TConfigEntry(ROUTER_GROUP, ROUTER_NAME_GRIDSTATCOLLECTINTERVAL,
                                           ROUTER_DEFAULT_GRIDSTATCOLLECTINTERVAL));
    Configuration->SetDefault(TConfigEntry(ROUTER_GROUP, ROUTER_NAME_MAXGRIDSTATENTRIES,
                                           ROUTER_DEFAULT_MAXGRIDSTATENTRIES));
    Configuration->SetDefault(TConfigEntry(ROUTER_GROUP, ROUTER_NAME_LOCALGRIDSTATOVERLAP,
                                           ROUTER_DEFAULT_LOCALGRIDSTATOVERLAP));
    Configuration->SetDefault(TConfigEntry(ROUTER_GROUP, ROUTER_NAME_COLLECTROUTESBYORIGIN,
                                           ROUTER_DEFAULT_COLLECTROUTESBYORIGIN));
    Configuration->SetDefault(TConfigEntry(ROUTER_GROUP, ROUTER_NAME_ROUTEORIGINSECTORS,
                                           ROUTER_DEFAULT_ROUTEORIGINSECTORS));
    Configuration->SetDefault(TConfigEntry(ROUTER_GROUP, ROUTER_NAME_TRAVELTIMEESTIMATEUPDATE,
                                           ROUTER_DEFAULT_TRAVELTIMEESTIMATEUPDATE));
    Configuration->SetDefault(TConfigEntry(ROUTER_GROUP, ROUTER_NAME_BINARYROUTEPLAN,
                                           ROUTER_DEFAULT_BINARYROUTEPLAN));
    ObjectManager->RegisterClass(CI_FLAG_TABLE, TRouteplan::ClassInfo,
                                 0, CLASS_ROUTEPLAN, this);
    ObjectManager->RegisterClass(0, TRoutingStat::ClassInfo,
                                 0, CLASS_ROUTINGSTAT, this);
    ObjectManager->RegisterClass(0, GridStatLookupClassInfo,
                                 CLASS_EDGE, CLASS_GRID_STAT_LOOKUP, this);
    ObjectManager->RegisterClass(CI_FLAG_TABLE, LocalGridStatLookupClassInfo,
                                 CLASS_EDGE, CLASS_LOCAL_GRID_STAT_LOOKUP, this);
    if (TMapStat::ClassInfo)
        TMapStat::ClassInfo->RegisterActionHandler(this);
    if (Parallel)
        Communication->RegisterRecipient(MTAG_ROUTER, this);
    return TRUE;
}

// =================================================================================== 

bool
TRouter::ActivateLocalRerouting(void)

{
    TravelTimeEstimateUpdateInterval = 
       Configuration->GetLong(ROUTER_GROUP, ROUTER_NAME_TRAVELTIMEESTIMATEUPDATE);    
    if (TravelTimeEstimateUpdateInterval > 0)
    {
        char    FilenamePattern[512];
        TString Filename;
        
        LocalSP = new TShortestPath(StartTimeStep, LocalGridStatLookupClassInfo);
        ReroutingLogActive = 
            Configuration->GetLong(ROUTER_GROUP, ROUTER_NAME_REROUTINGLOGACTIVE);
        ReroutingFraction = 
            Configuration->GetDouble(ROUTER_GROUP, ROUTER_NAME_REROUTINGFRACTION);      
        ReroutingMinRelativeDelay = 
            Configuration->GetDouble(ROUTER_GROUP, ROUTER_NAME_REROUTINGMINRELDELAY);   
        ReroutingMinRelativeImprovement = 
            Configuration->GetDouble(ROUTER_GROUP, ROUTER_NAME_REROUTINGMINRELIMPROVEMENT);
        ReroutingFractionLimit = (long) (1000 * ReroutingFraction);
        ReroutingLog = new TStream;
        sprintf(FilenamePattern, "%s.%%P%s", RouteplanBasename(), REROUTING_LOGFILE_EXTENSION);
        Filename = Environment->ReplacePathTokens(FilenamePattern);
        if (!ReroutingLog->Open(Filename, "w"))
        {
            Error("ActivateStatistics: cannot open re-routing log file %s", Filename());
            return FALSE;
        }
        Configuration->WriteToFile(*ReroutingLog);
        Print("opened local re-routing stat file %s", Filename());
        
    }    
    return TRUE;
}

// =================================================================================== 

bool
TRouter::ActivateStatistics(void)

{
    if (Parallel->IsMaster())
    {
        RoutingStatInterval = (int) Configuration->GetLong(ROUTER_GROUP,
                                                           ROUTER_NAME_ROUTINGSTATINTERVAL);
        RoutingStatDump = (int) Configuration->GetLong(ROUTER_GROUP,
                                                       ROUTER_NAME_ROUTINGSTATDUMP);
        
        // start routing statistics:
        
        if (RoutingStatInterval > 0)
        {
            RoutingSamplingDescription = 
                 new TSamplingDescription(TRoutingStat::ClassInfo->GetID(), SF_AGGREGATE,
                                          Parallel->GetLocalCPNID(), SAMPLING_ID_ROUTING_STAT);
            RoutingAggregationInfo = new TAggregationInfo(this, RoutingSamplingDescription);
            RoutingSamplingDescription->SetCollectInterval(RoutingStatInterval);
            StatManager->RegisterAggregation(RoutingAggregationInfo);
            StatManager->StartSampling(RoutingSamplingDescription);
            if (Configuration->Exists(ROUTER_GROUP, ROUTER_NAME_ROUTINGSTATFILE))
            {
                char Filename[512];
                
                RoutingStatOutput = new TStream;
                sprintf(Filename, "%s%s", RouteplanBasename(), ROUTING_STATFILE_EXTENSION);
                if (!RoutingStatOutput->Open(Filename, "w"))
                {
                    Error("ActivateStatistics: cannot open routing stat file %s", Filename);
                    return FALSE;
                }
                TRoutingStat::WriteStatDescription(RoutingStatOutput);
                Configuration->WriteToFile(*RoutingStatOutput);
            }
        }       
    }

    // start grid statistics:

    GridStatSampleInterval = Configuration->GetLong(ROUTER_GROUP, 
                                                    ROUTER_NAME_GRIDSTATSAMPLEINTERVAL);
    GridStatCollectInterval = Configuration->GetLong(ROUTER_GROUP, 
                                                     ROUTER_NAME_GRIDSTATCOLLECTINTERVAL);
    MaxGridStatEntries = Configuration->GetLong(ROUTER_GROUP, ROUTER_NAME_MAXGRIDSTATENTRIES);
    if (GridStatCollectInterval > 0)
    {
        if (Parallel->IsMaster())
        {
            GridSamplingDescription = 
                  new TSamplingDescription(TGridStat::ClassInfo->GetID(), SF_MERGE,
                                           Parallel->GetLocalCPNID(), SAMPLING_ID_GRID_STAT);
            GridAggregationInfo = new TAggregationInfo(this, GridSamplingDescription);
            GridSamplingDescription->SetSampleInterval(GridStatSampleInterval);
            GridSamplingDescription->SetCollectInterval(GridStatCollectInterval);
            StatManager->RegisterAggregation(GridAggregationInfo);
            StatManager->StartSampling(GridSamplingDescription);
        }
        LocalGridStatOverlap = Configuration->GetDouble(ROUTER_GROUP,
                                                        ROUTER_NAME_LOCALGRIDSTATOVERLAP);
        if (LocalGridStatOverlap > 0.0)
        {
            LocalGridSamplingDescription = 
                   new TSamplingDescription(TGridStat::ClassInfo->GetID(), SF_MERGE,
                                            DEST_LIST, SAMPLING_ID_LOCAL_GRID_STAT);
            LocalGridAggregationInfo = new TAggregationInfo(this, LocalGridSamplingDescription);
            LocalGridSamplingDescription->SetDestList(&LocalGridDestList);
            LocalGridSamplingDescription->SetSampleInterval(GridStatSampleInterval);
            LocalGridSamplingDescription->SetCollectInterval(GridStatCollectInterval);
            StatManager->RegisterAggregation(LocalGridAggregationInfo);
            StatManager->StartSampling(LocalGridSamplingDescription);
        }
    }
    return TRUE;
}

// ===================================================================================

bool
TRouter::AreRoutesPending(void) const

{
    return RouteplanParser AND !RouteplanParser->EndOfFile();
}

// ===================================================================================

bool
TRouter::CheckRouteplans(TTimeStep TimeStep)

{
    long PlanTimeStep = 0;

    while (RouteplanParser AND !RouteplanParser->EndOfFile() AND
           TimeStep >= CurrentRouteplanTimeStep)
    {
        if (!ReadRoute(CurrentRouteplanTimeStep))
            return FALSE;
        if (RouteplanParser->GetTokenID(TOKEN_ROUTE) == TOKEN_ROUTE AND !BinaryRouteplan)
            if (!RouteplanParser->ReadToken()) // skip token ROUTE
            {
                Error("CheckRouteplans: cannot read token ROUTE");
                return FALSE;
            }
        if (RouteplanParser->EndOfFile())
        {
            CloseRouteplanInput();
            return TRUE;
        }
        if (!RouteplanParser->Decode(PlanTimeStep))
        {
            if (RouteplanParser->EndOfFile())
            {
                CloseRouteplanInput();
                return TRUE;
            }
            Error("CheckRouteplans: cannot decode time-stamp");
            return FALSE;
        }
        if (PlanTimeStep < CurrentRouteplanTimeStep)
        {
            Error("error in route plan timesteps: last:%ld current:%ld",
                  CurrentRouteplanTimeStep, (long) PlanTimeStep);
            return FALSE;
        }
        CurrentRouteplanTimeStep = PlanTimeStep;
    }
    return TRUE;
}

/* =================================================================================== */

bool
TRouter::CloseFiles(void)

{
    bool Return = TRUE;

    if (RouteplanOutput)
    {
        if (SortByDepartureTime)
            if (!SortRouteplans())
            {      
                Error("CloseFiles: cannot sort routeplans");
                Return = FALSE;
            }
        if (RouteplanTempOutput)
        {
            DeleteObject(RouteplanTempOutput);
            unlink(RouteplanTempFilename);
        }
        if (OffsetFile)
        {
            fclose(OffsetFile);
            OffsetFile = 0;
            unlink(OffsetFilename);
        }
        if (RouteplanOutput)
        {
            Print("closing routeplan file");
            DeleteObject(RouteplanOutput);
        }
        if (ConversionLogOutput)
        {
            Print("closing conversion logfile with %ld entries", ConversionLogEntries);
            fclose(ConversionLogOutput);
            ConversionLogOutput = 0;
        }    
        if (ConversionStatOutput)
        {
            Print("closing conversion stat file");
            fclose(ConversionStatOutput);
            ConversionStatOutput = 0;
        }    
    }
    if (RoutingStatOutput)
    {
        Print("closing routing stat file");
        DeleteObject(RoutingStatOutput);
    }    
    if (CompletedRoutesOutput)
    {
        Print("closing completed routes output");
        DeleteObject(CompletedRoutesOutput);
    }
    if (RoutesByOriginOutput)
    {
        Print("closing routes by originoutput");
        DeleteObject(RoutesByOriginOutput);
    }
    if (TransimsLinkStatOutput)
    {
        Print("closing TRANSIMS link stat output");
        DeleteObject(TransimsLinkStatOutput);
    }
    if (ReroutingLog)
    {
        Print("closing re-routing log");
        DeleteObject(ReroutingLog);
    }
    return Return;
}

// =================================================================================== 

bool
TRouter::CloseRouteplanInput(void)

{
    if (RouteplanParser)
    {
        DeleteObject(RouteplanParser);
        Print("closing routeplan input");
    }
    return TRUE;
}

// =================================================================================== 

bool
TRouter::ClipTransimsRoute2Pamina(LongID PlanID, long MinTime, long MaxTime,
                                  LongID SourceAccessory, LongID DestAccessory,
                                  TRouteStep *TransimsEntries, int NrOfTransimsLinks,
                                  TEdge **PaminaLinks, int &NrOfPaminaLinks,
                                  int &FirstPaminaLink, short int &RouteplanFlags,
                                  LongID &SourceID, LongID &DestID)

{
    bool         Found;
    bool         SingleEdgeFound;
    int          SourceLink;
    TParallelID  PaminaLinkID;
    TParallelID  CommonNodeID;
    TEdge       *FromToEdge, *ToFromEdge;
    TEdge       *NewFromToEdge = 0, *NewToFromEdge = 0, *NewEdge;
    TParallelID  FirstNodeID1, FirstNodeID2, SecondNodeID1, SecondNodeID2;
    long         DepartureTime;

    ErrorID = ROUTE_OK;
    RouteplanFlags = 0;
    RoutesProcessed++;
    if (NrOfTransimsLinks == 0)
    {
        ErrorID = ROUTE_TOO_SHORT;
        WriteLog("plan %ld is too short", PlanID);
        return FALSE;
    }
    NrOfPaminaLinks = 0;
    SourceLink = 0;
    PaminaLinkID = 2 * TransimsEntries[SourceLink].ID;
    FromToEdge = (TEdge *) TEdge::ClassInfo->GetObjectPtr(PaminaLinkID);
    ToFromEdge = (TEdge *) TEdge::ClassInfo->GetObjectPtr(PaminaLinkID + 1);
    Found = SingleEdgeFound = FALSE;
    SourceLink++;
    while (SourceLink < NrOfTransimsLinks - 1 AND !Found)
    {
        if (FromToEdge != 0 OR ToFromEdge != 0)
            SingleEdgeFound = TRUE;
        PaminaLinkID = 2 * TransimsEntries[SourceLink].ID;
        NewFromToEdge = (TEdge *) TEdge::ClassInfo->GetObjectPtr(PaminaLinkID);
        NewToFromEdge = (TEdge *) TEdge::ClassInfo->GetObjectPtr(PaminaLinkID + 1);
        if ((FromToEdge != 0 OR ToFromEdge !=0) AND
            (NewFromToEdge !=0 OR NewToFromEdge != 0))
            Found = TRUE;
        else
        {
            FromToEdge = NewFromToEdge;
            ToFromEdge = NewToFromEdge;
        }
        SourceLink++;
    }
    if (!Found)
    {
        if (SingleEdgeFound)
        {
            ErrorID = ROUTE_TOO_SHORT;
            WriteLog("plan %ld has a single edge (not handled yet)", PlanID);
        }
        else
        {
            WriteLog("plan %ld does not touch map", PlanID);
            RoutesOutsideRegion++;
        }
        return FALSE;
    }    
    FirstPaminaLink = SourceLink - 2;
    DepartureTime = TransimsEntries[FirstPaminaLink].Estimate;
    if (DepartureTime < MinTime OR DepartureTime > MaxTime)
    {
        WriteLog("plan #%ld (time %d:%02d) outside time window (%d:%02d - %d:%02d)", 
                 PlanID, DepartureTime / 3600, DepartureTime / 60 % 60,
                 MinTime / 3600, MinTime / 60 % 60,
                 MaxTime / 3600, MaxTime / 60 % 60);
        RoutesOutsideTimeWindow++;
        return FALSE;
    }

    // This can probably be done in a much shorter fashion:

    if (FromToEdge)
    {
        FirstNodeID1 = FromToEdge->GetFromNodeID();
        FirstNodeID2 = FromToEdge->GetToNodeID();
    }
    else
    {
        FirstNodeID1 = ToFromEdge->GetFromNodeID();
        FirstNodeID2 = ToFromEdge->GetToNodeID();
    }
    if (NewFromToEdge)
    {
        SecondNodeID1 = NewFromToEdge->GetFromNodeID();
        SecondNodeID2 = NewFromToEdge->GetToNodeID();
    }
    else
    {
        SecondNodeID1 = NewToFromEdge->GetFromNodeID();
        SecondNodeID2 = NewToFromEdge->GetToNodeID();
    }

    //    Print(" %lld %lld -> %lld %lld", FirstNodeID1, FirstNodeID2, SecondNodeID1, SecondNodeID2);

    if (FirstNodeID1 == SecondNodeID1 OR FirstNodeID1 == SecondNodeID2)
    {
        PaminaLinks[0] = FromToEdge;
        PaminaLinks[1] = NewToFromEdge;
        CommonNodeID = FirstNodeID1;
    } 
    else if (FirstNodeID2 == SecondNodeID1 OR FirstNodeID2 == SecondNodeID2)
    {
        PaminaLinks[0] = ToFromEdge;
        PaminaLinks[1] = NewToFromEdge;
        CommonNodeID = FirstNodeID2;
    } 
    else
    {
        WriteLog("no common initial node found for plan %ld", PlanID);
        RoutesCorrupted++;
        return FALSE;
    }
    if (FromToEdge != 0 AND FromToEdge->GetToNodeID() == CommonNodeID)
        PaminaLinks[0] = FromToEdge;
    else
        if (ToFromEdge != 0 AND ToFromEdge->GetToNodeID() == CommonNodeID)
            PaminaLinks[0] = ToFromEdge;
        else
        {
            WriteLog("inconsistent first edge in plan %ld", PlanID);

	    //	    Print("%s %s Common=%lld", 
	    //		  (FromToEdge == 0)? "FromToEdge=0":(char *) FromToEdge->GetInstanceName(),
	    //		  (ToFromEdge == 0)? "ToFromEdge=0":(char *) ToFromEdge->GetInstanceName(),
	    //		  CommonNodeID);


            RoutesCorrupted++;
            return FALSE;
        }
    if (NewFromToEdge != 0 AND NewFromToEdge->GetFromNodeID() == CommonNodeID)
        PaminaLinks[1] = NewFromToEdge;
    else
        if (NewToFromEdge != 0 AND NewToFromEdge->GetFromNodeID() == CommonNodeID)
            PaminaLinks[1] = NewToFromEdge;
        else
        {
            WriteLog("inconsistent second edge in plan %ld", PlanID);
            RoutesCorrupted++;
            return FALSE;
        }
    CommonNodeID = PaminaLinks[1]->GetToNodeID();
    NrOfPaminaLinks = 2;
    while (SourceLink < NrOfTransimsLinks)
    {
        PaminaLinkID = 2 * TransimsEntries[SourceLink].ID;
        FromToEdge = (TEdge *) TEdge::ClassInfo->GetObjectPtr(PaminaLinkID);
        ToFromEdge = (TEdge *) TEdge::ClassInfo->GetObjectPtr(PaminaLinkID + 1);
        if (FromToEdge == 0 AND ToFromEdge == 0)
            break;
        if (FromToEdge AND FromToEdge->GetFromNodeID() == CommonNodeID)
            NewEdge = FromToEdge;
        else
            if (ToFromEdge AND ToFromEdge->GetFromNodeID() == CommonNodeID)
                NewEdge = ToFromEdge;
            else
            {
                WriteLog("no successor node found for %ld in plan %ld",
                         CommonNodeID, PlanID);
                RoutesCorrupted++;
                return FALSE;
            }
        PaminaLinks[NrOfPaminaLinks] = NewEdge;
        NrOfPaminaLinks++;
        CommonNodeID = NewEdge->GetToNodeID();
        SourceLink++;
    }
    if (FirstPaminaLink == 0)
    {
        RouteplanFlags |= RP_FLAG_SOURCE_IS_ACCESSORY;
        if (!GetAccessory(PaminaLinks[0], SourceAccessory, SourceID))
        {
            WriteLog("cannot find source accessory %ld in plan %ld", SourceAccessory, PlanID);
            return FALSE;
        }
    }
    else
    {
        RoutesStartingOutsideRegion++;
        SourceID = PaminaLinks[0]->GetFromNodeID();
    }
    if (FirstPaminaLink + NrOfPaminaLinks == NrOfTransimsLinks)
    {
        RouteplanFlags |= RP_FLAG_DEST_IS_ACCESSORY;
        if (!GetAccessory(PaminaLinks[NrOfPaminaLinks - 1], DestAccessory, DestID))
        {
            WriteLog("cannot find destination accessory %ld in plan %ld", DestAccessory, PlanID);
            return FALSE;
        }
    }
    else
    {
        RoutesEndingOutsideRegion++;
        DestID = PaminaLinks[NrOfPaminaLinks - 1]->GetToNodeID();
    }
    RoutesValid++;
    if (NrOfPaminaLinks < MinRouteLength)
        MinRouteLength = NrOfPaminaLinks;
    if (NrOfPaminaLinks > MaxRouteLength)
        MaxRouteLength = NrOfPaminaLinks;
    SumRouteLength += NrOfPaminaLinks;
    return TRUE;
}


// =================================================================================== 

bool
TRouter::Collect(TSamplingDescription *Description)

{
    if (Description->GetStatObjectClassID() == TRoutingStat::ClassInfo->GetID())
    {
        TRoutingStat *Stat = new TRoutingStat;

        Stat->SetTimeStep(Description->GetTimeStep());
        if (!CollectRoutingStat(Stat))
        {
            Error("Collect: cannot collect routing stat");
            return FALSE;
        }
        Description->GetObjectList()->Append(Stat);
        return TRUE;
    }
    return TStatHandler::Collect(Description);
}

// ===================================================================================

bool
TRouter::CollectRoutingStat(TRoutingStat *Stat)

{
    TEnumerator < TSource > CSource(TSource::ClassInfo->GetTable());
    TEnumerator < TSink > CSink(TSink::ClassInfo->GetTable());

    CSource.Enumerate();
    while (CSource())
    {
        Stat->Collect(CSource);
        CSource++;
    }
    CSink.Enumerate();
    while (CSink())
    {
        Stat->Collect(CSink);
        CSink++;
    }
    if (AreRoutesPending())
        Stat->CountRouteplanFile();
    return TRUE;
}

// =================================================================================== 

bool
TRouter::EstimateRemainingTravelTimes(const TTimeStep &CurrentTime)

{
    TEnumerator< TMultilaneGrid > CGrid(TMultilaneGrid::ClassInfo->GetTable());

    CGrid.Enumerate();
    while (CGrid())
    {
        TEdge      *Edge = 0;
        long        CLane, CSite;
        TVehicle   *Vehicle;
        TRouteplan *Route = 0;
        int         LastValidStep;

        if (TimeEdges)
        {
            if ((Edge = TEdge::GetPtr(CGrid->GetParallelID())) == 0)
            {
                Error("ExecuteAccessories: cannot retrieve edge");
                return FALSE;
            }
            if (!Edge->StartTimer(EdgeTimerClassInfo->GetInfoID()))
            {
                Error("ExecuteAccessories: cannot start edge timer");
                return FALSE;
            }
        }
        for (CSite = CGrid->GetFirstSite(); CSite < CGrid->GetLastSite(); CSite++)
            for (CLane = 0; CLane < CGrid->GetNrOfLanes(); CLane++)
                if ((Vehicle = CGrid->VAccess(CLane, CSite)) AND
                    (Route = Vehicle->GetRouteplan()) AND
                     IsInReroutingSelection(Route))
                {
                    long   EstimatedArrivalTimeStep;
                    double RelativeLocation;
                    
                    RelativeLocation = 1.0 * CSite / CGrid->GetNrOfSites();
                    if (!LocalSP->EstimateRemainingTravelTime(CurrentTime, CGrid->GetParallelID(),
                                                              RelativeLocation, Route,
                                                              LastValidStep, 
                                                              EstimatedArrivalTimeStep))
                    {
                        Error("cannot estimate travel time for %s",
                              (char *) Route->GetObjectName());
                        return FALSE;
                    }
                    if (Route->GetEstimatedRelativeDelay() > ReroutingMinRelativeDelay)
                        if (!Reroute(CurrentTime, CGrid->GetParallelID(),
                                     RelativeLocation, Vehicle, LastValidStep,
                                     EstimatedArrivalTimeStep))
                            {
                            Error("EstimateRemainingTravelTimes: cannot re-route %s",
                                  (char *) Route->GetObjectName());
                            return FALSE;
                        }
                }
        if (TimeEdges)
            Edge->StopTimer(EdgeTimerClassInfo->GetInfoID());
        CGrid++;
    }
    return TRUE;
}

// ===================================================================================

bool
TRouter::ExecuteSubTimeStep(const TTimeStep &TimeStep)

{
    if (TimeStep.GetSubStep() == 0)
    {
        if (!CheckRouteplans(TimeStep))
        {
            Error("ExecuteSubTimeStep: while checking for departing routes");
            return FALSE;
        }
        if (!SendExecutedRouteplans())
        {
            Error("ExecuteSubTimeStep: cannot send executed routeplans");
            return FALSE;
        }
        if (LocalSP AND TravelTimeEstimateUpdateInterval > 0 AND LocalGridStatValid)
            if ((long) TimeStep % TravelTimeEstimateUpdateInterval == 0)
                if (!EstimateRemainingTravelTimes(TimeStep))
                {
                    Error("ExecuteSubTimeStep: cannot estimate remaining travel times");
                    return FALSE;
                }
    }
    return TRUE;
}

// ===================================================================================

bool
TRouter::GetAccessory(TEdge *Edge, LongID TransimsAccessoryID, LongID &PaminaAccessoryID)

{
    if (Edge->GetParallelID() & 1)
        PaminaAccessoryID = 2 * TransimsAccessoryID + 1;
    else
        PaminaAccessoryID = 2 * TransimsAccessoryID;
    return TRUE;
}

// ===================================================================================

TString
TRouter::GetClassName(void) const

{
    return "TRouter";
}

// ===================================================================================

double
TRouter::GetCompletion(void)

{
    if (LatestRoutingStat == 0)
        return 0.0;
    if (LatestRoutingStat->GetOpenRouteplanFiles() > 0)
        return 0.0;
    if (LatestRoutingStat->GetInsertions() == 0)
        return 0.0;
    return (double) (LatestRoutingStat->GetDeletions() + LatestRoutingStat->GetRoutesFailed()) / 
        LatestRoutingStat->GetInsertions();
}

// ===================================================================================

TString
TRouter::GetErrorString(void) const

{
    char line[256];

    sprintf(line, RouteErrors[ErrorID], ErrorEdgeID, SecondErrorEdgeID);
    return line;
}

// =================================================================================== 

bool
TRouter::GetLocation(TParallelID ID, bool IsAccessory, TLocation &Location)

{
    if (IsAccessory)
    {
        //      TAccessory *Accessory;
        TEdge      *Edge;
        TNode      *FromNode;
        TNode      *ToNode;
        TLocation   Vector;
        TInfoID     InfoID;

        /*
        if ((Accessory = TAccessory::GetPtr(ID)) == 0)
        {
            Error("GetLocation: cannot retrieve accessory %ld", ID);
            return FALSE;
        }
        if ((Edge = TEdge::GetPtr(Accessory->GetEdgeID())) == 0)
        {
            Error("GetLocation: cannot retrieve edge for %s", 
                  (char *) Accessory->GetObjectName());
            return FALSE;
        }
        */
        if ((Edge = TEdge::GetPtr(ID)) == 0)
        {
            Error("GetLocation: cannot retrieve edge %ld", ID);
            return FALSE;
        }
        InfoID = TLocationInfo::ClassInfo->GetInfoID();
        FromNode = Edge->GetFromNode();
        ToNode = Edge->GetToNode();
        Vector = ToNode->Loc(InfoID) - FromNode->Loc(InfoID);
        //      Location = FromNode->Loc(InfoID) + Accessory->GetRelativeLocation() * Vector;
        Location = FromNode->Loc(InfoID) + 0.5 * Vector;
    }
    else
    {
        TNode *Node;

        if ((Node = TNode::GetPtr(ID)) == 0)
        {
            Error("GetLocation: cannot get node for ID %ld", ID);
            return FALSE;
        }
        Location = Node->Loc(TLocationInfo::ClassInfo->GetInfoID());
    }
    return TRUE;
}

// =================================================================================== 

bool
TRouter::HandleAggregationResult(TAggregationInfo *AggregationInfo, 
                                 TTimeStep &TimeStep)

{
    if (AggregationInfo == GridAggregationInfo OR 
        AggregationInfo == LocalGridAggregationInfo)
        return HandleGridResult(AggregationInfo, TimeStep);
    if (AggregationInfo == RoutingAggregationInfo)
        return HandleRoutingResult(AggregationInfo, TimeStep);
    return TStatHandler::HandleAggregationResult(AggregationInfo, TimeStep);
}

// =================================================================================== 

bool
TRouter::HandleAnonymousAction(TClassInfo *ClassInfo, TObjectAction Action)

{
    if (ClassInfo == TMapStat::ClassInfo)
    {
        if (!UpdateGridStatDestList())
        {
            Error("HandleAnonymousAction: cannot update grid stat destinations");
            return FALSE;
        }
    }
    else
        return TObjectHandler::HandleAnonymousAction(ClassInfo, Action);
    return TRUE;
}

// =================================================================================== 

bool
TRouter::HandleCreateRequest(TParallelObject *&Object, const TClassInfo *ClassInfo, 
                             TParallelID ParallelID)

{
    if (ClassInfo == TRouteplan::ClassInfo)

        Object = new TRouteplan(ParallelID);

    if (ClassInfo == TRoutingStat::ClassInfo)

        Object = new TRoutingStat(ParallelID);

    if (ClassInfo == GridStatLookupClassInfo)

        Object = new TGridStatLookup(GridStatLookupClassInfo, ParallelID);

    if (ClassInfo == LocalGridStatLookupClassInfo)

        Object = new TGridStatLookup(LocalGridStatLookupClassInfo, ParallelID);

    return TRUE;
}

// =================================================================================== 

bool
TRouter::HandleEvent(TEventID EventID, TBaseMessage *Message, int &Handled)

{
    switch(EventID)
    {
    case MEVENT_ROUTER_COMPLETED_ROUTE:
        Handled = TRUE;
        return HandleEventCompletedRoute(Message);
        
    default:
        Error("illegal event ID %ld", EventID);
        return FALSE;
    }
#ifndef COMPLETE_SWITCH_FLOW
    return TRUE;
#endif
}

/* =================================================================================== */

bool
TRouter::HandleEventCompletedRoute(TBaseMessage *Message)

{
    TRouteplan    *Routeplan = 0;
    TRouteplanList Routes;
    TRouteplanEnum CRoute;

    if ((Routeplan = (TRouteplan *) ObjectManager->ReceiveObjectList((TParallelObjectList *) &Routes, 
                                                                     Message, TRUE)) == 0)
    {
        Error("HandleEventCompletedRoute: cannot receive");
        return FALSE;
    }
    CRoute.WalksForwardThrough(Routes);
    while (CRoute())
    {   
        if (!CRoute->WriteStat(CompletedRoutesOutput))
        {
            Error("HandleEventCompletedRoute: cannot write to file");
            return FALSE;
        }
        if (RoutesByOriginOutput)
            if (!WriteRouteOriginStat(CRoute))
            {
                Error("HandleEventCompletedRoute: cannot output origin stat for route %s",
                      (char *) CRoute->GetInstanceName());
                return FALSE;
            }
        CRoute++;
    }
    Routes.DeleteAll(TRUE);
    return TRUE;
}

// =================================================================================== 

bool
TRouter::HandleGridResult(TAggregationInfo *AggregationInfo, TTimeStep &TimeStep)

{
    TAggregationResult *Result = 0;
    TGridStatList      *StatList = 0;

    if ((Result = (TAggregationResult *) AggregationInfo->GetResult(TimeStep)) == 0)
    {
        Error("HandleGridResult: cannot retrieve aggregation result");
        return FALSE;
    }
    if ((StatList = (TGridStatList *) Result->StatObjectList) == 0)
    {
        Error("HandleGridResult: cannot retrieve stat list");
        return FALSE;
    }
    if (AggregationInfo == GridAggregationInfo)
    {
        if (!WriteTransimsLinkStat(StatList))
        {
            Error("HandleGridResult: while writing TRANSIMS link stat");
            return FALSE;
        }    
        if (!StoreGridStats(GridStatLookupClassInfo, StatList))
        {
            Error("handleGridResult: while storing grid results");
            return FALSE;
        }
    }
    else
    {
        if (!StoreGridStats(LocalGridStatLookupClassInfo, StatList))
        {
            Error("handleGridResult: while storing local grid results");
            return FALSE;
        }
        if (Result->NrOfCPN > 1 AND !LocalGridStatValid)
        {
            Print("non-trivial local grid statistics received -> activate re-routing");
            LocalGridStatValid = TRUE;
        }
    }
    AggregationInfo->DisposeResult(TimeStep);
    return TRUE;
}

/* =================================================================================== */

bool
TRouter::HandleExecutedRouteplan(const TTimeStep &TimeStep, TRouteplan *Routeplan)

{
    Routeplan->SetDeletionTimeStep(StartTimeStep + (long) TimeStep);
    ExecutedRouteplans.Append(Routeplan);
    return TRUE;
}

// =================================================================================== 

bool
TRouter::HandleLoadRequest(TParallelObject *Object, bool &Loaded)

{
    if (Object->GetClassID() == GridStatLookupClassInfo->GetID() OR
        Object->GetClassID() == LocalGridStatLookupClassInfo->GetID())
    {
        Loaded = TRUE;
    }
    return TRUE;
}

// =================================================================================== 

bool
TRouter::HandleRoutingResult(TAggregationInfo *Info, TTimeStep &TimeStep)
    
{
    TAggregationResult *Result = 0;
    TRoutingStat       *Stat = 0;

    if ((Result = (TAggregationResult *) Info->GetResult(TimeStep)) == 0)
    {
        Error("HandleRoutingResult: cannot retrieve aggregation result");
        return FALSE;
    }
    if ((Stat = (TRoutingStat *) Result->StatObject) == 0)
    {
        Error("HandleRoutingResult: cannot retrieve stat object");
        return FALSE;
    }
    DeleteObject(LatestRoutingStat);
    if (RoutingStatDump)
        Stat->PrintScreenStat(stderr);
    if (RoutingStatOutput)
    {
        Stat->Write(StartTimeStep, RoutingStatOutput);
        RoutingStatOutput->Flush();
    }
    LatestRoutingStat = Stat;
    Result->StatObject = 0;
    Info->DisposeResult(TimeStep);
    return TRUE;
}

/* =================================================================================== */

bool
TRouter::HandleNewRouteplan(const TTimeStep &TimeStep, TRouteplan *Routeplan)

{
    Routeplan->SetInsertionTimeStep(StartTimeStep + (long) TimeStep);
    Routeplan->Step();
    return TRUE;
}

/* =================================================================================== */

bool
TRouter::InitRoutesByOrigin(void)

{
    if (Configuration->GetLong(ROUTER_GROUP, ROUTER_NAME_COLLECTROUTESBYORIGIN))
    {
        if (!OpenRoutesByOriginOutput())
        {
            Error("InitRoutesByRogin: cannot open output");
            return FALSE;
        }
        RouteOriginSectors = Configuration->GetLong(ROUTER_GROUP, 
                                                    ROUTER_NAME_ROUTEORIGINSECTORS);
        RouteDestRectangle = 
            TRectangle(TLocation(Configuration->GetDouble(ROUTER_GROUP, 
                                                          ROUTER_NAME_ROUTEDESTLOWERLEFTX),
                                 Configuration->GetDouble(ROUTER_GROUP, 
                                                          ROUTER_NAME_ROUTEDESTLOWERLEFTY)),
                       TLocation(Configuration->GetDouble(ROUTER_GROUP, 
                                                          ROUTER_NAME_ROUTEDESTUPPERRIGHTX),
                                 Configuration->GetDouble(ROUTER_GROUP, 
                                                          ROUTER_NAME_ROUTEDESTUPPERRIGHTY)));
        RouteDestLocation = RouteDestRectangle.Center();
        Print("route destination set to %s", (char *) RouteDestRectangle.GetInstanceName());
    }
    return TRUE;
}

/* =================================================================================== */

bool
TRouter::IsInReroutingSelection(TRouteplan *Route)

{
    int RouteID;
    
    RouteID = Route->GetRouteID() % 1000;
    return RouteID < ReroutingFractionLimit;
}

/* =================================================================================== */

bool
TRouter::OpenConversionLogFile(void)

{
    if (ConversionLogOutput == 0)
    {
        char Filename[512];

        sprintf(Filename, "%s%s", RouteplanBasename(), CONVERSION_LOGFILE_EXTENSION);
        if ((ConversionLogOutput = fopen(Filename, "w")) == 0)
        {
            Error("cannot open conversion logfile %s for output", Filename);
            return FALSE;
        }
        Print("conversion logfile %s opened", Filename);
    }
    return TRUE;
}

/* =================================================================================== */

bool
TRouter::OpenConversionStatFile(void)

{
    if (ConversionStatOutput == 0)
    {
        char Filename[512];

        sprintf(Filename, "%s%s", RouteplanBasename(), CONVERSION_STATFILE_EXTENSION);
        if ((ConversionStatOutput = fopen(Filename, "w")) == 0)
        {
            Error("cannot open conversion stat stat file %s for output", Filename);
            return FALSE;
        }
        Print("conversion stat file %s opened", Filename);
    }
    return TRUE;
}

/* =================================================================================== */

bool
TRouter::OpenCompletedRoutesOutput(void)

{
    if (CompletedRoutesOutput == 0)
    {
        char Filename[512];

        CompletedRoutesOutput = new TStream;
        sprintf(Filename, "%s%s", RouteplanBasename(), COMPLETED_ROUTES_EXTENSION);     
        if (!CompletedRoutesOutput->Open(Filename, "w"))
        {
            Error("cannot open completed routes file %s", Filename);
            return FALSE;
        }
        Print("completed routes file %s opened", Filename);
    }
    return TRUE;
}

/* =================================================================================== */

bool
TRouter::OpenRouteplanForReading(void)

{
    if (!Configuration->Exists(ROUTER_GROUP, ROUTER_NAME_ROUTEPLAN))
    {
        Error("OpenRouteplanForReading: no routeplan filename given");
        return FALSE;
    }
    RouteplanParser = new TParser(NR_OF_ROUTE_PLAN_TOKENS, RouteplanTokens);
    RouteplanBasename = Environment->ReplacePathTokens(Configuration->GetString());
    Environment->SetRunName(RouteplanBasename);
    RouteplanFilename = RouteplanBasename + 
                        Configuration->GetString(ROUTER_GROUP, ROUTER_NAME_ROUTEPLANEXTENSION);
    BinaryRouteplan = Configuration->GetLong(ROUTER_GROUP, ROUTER_NAME_BINARYROUTEPLAN);
    if (!RouteplanParser->Open(RouteplanFilename, "r", BinaryRouteplan))
    {
        Error("OpenRouteplanForReading: cannot open '%s'", (char *) RouteplanFilename);
        return FALSE;
    }
    Print("routeplan file '%s' opened...", (char *) RouteplanFilename);

    // read in the first route header (if available, otherwise the file is empty!)

    if (BinaryRouteplan)
    {
        SkipRouteplan = new TRouteplan(INVALID);
        SkipVehicle = new TVehicle(INVALID);
    }
    else
        if (RouteplanParser->ReadToken())
            if (!(RouteplanParser->GetTokenID(TOKEN_ROUTE) == TOKEN_ROUTE AND
                  RouteplanParser->ReadToken()))
            {
                Error("OpenRouteplanForReading: cannot read first route header");
                return FALSE;
            }
    if (!RouteplanParser->Decode(CurrentRouteplanTimeStep))
    {
        Error("OpenRouteplanForReading: cannot read first route header");
        return FALSE;
    }
    return TRUE;
}

/* =================================================================================== */

bool
TRouter::OpenRouteplanForWriting(bool SortByDepartureTimeFlag)

{
    if (!Configuration->Exists(ROUTER_GROUP, ROUTER_NAME_ROUTEPLAN))
    {
        Error("OpenRouteplanForWriting: no routeplan filename given");
        return FALSE;
    }
    RouteplanBasename = Environment->ReplacePathTokens(Configuration->GetString());
    RouteplanFilename = RouteplanBasename + 
                        Configuration->GetString(ROUTER_GROUP, ROUTER_NAME_ROUTEPLANEXTENSION);
    BinaryRouteplan = Configuration->GetLong(ROUTER_GROUP, ROUTER_NAME_BINARYROUTEPLAN);
    SortByDepartureTime = SortByDepartureTimeFlag;
    if (SortByDepartureTime)
    {
        OffsetFilename = Environment->ReplacePathTokens(OFFSET_FILENAME);
        if ((OffsetFile = fopen(OffsetFilename, "w+")) == 0)
        {
            Error("cannot open offset file '%s' for writing", OffsetFilename());
            return FALSE;
        }
        RouteplanTempFilename = Environment->ReplacePathTokens(ROUTEPLAN_FILENAME);
        RouteplanTempOutput = new TStream;
        if (!RouteplanTempOutput->Open(RouteplanTempFilename, "w+", BinaryRouteplan))
        {
            Error("cannot open temporary routeplan file '%s' for writing", 
                  RouteplanTempFilename());
            return FALSE;
        }
    }
    RouteplanOutput = new TStream;
    if (!RouteplanOutput->Open(RouteplanFilename, "w"))
    {
        Error("cannot open routeplan file '%s' for writing", RouteplanFilename());
        return FALSE;
    }
    return TRUE;
}

/* =================================================================================== */

bool
TRouter::OpenRoutesByOriginOutput(void)

{
    if (RoutesByOriginOutput == 0)
    {
        char Filename[512];

        RoutesByOriginOutput = new TStream;
        sprintf(Filename, "%s%s", RouteplanBasename(), ROUTES_BY_ORIGIN_EXTENSION);     
        if (!RoutesByOriginOutput->Open(Filename, "w"))
        {
            Error("cannot open routes by origin file %s", Filename);
            return FALSE;
        }
        Print("routes by origin file %s opened", Filename);
    }
    return TRUE;
}

/* =================================================================================== */

bool
TRouter::OpenTransimsLinkStatOutput(void)

{
    if (TransimsLinkStatOutput == 0 AND
        Configuration->Exists(ROUTER_GROUP, ROUTER_NAME_TRANSIMSLINKSTAT))
    {
        char Filename[512];
        
        TransimsLinkStatOutput = new TStream;
        sprintf(Filename, "%s%s", RouteplanBasename(), TRANSIMS_LINK_STAT_EXTENSION);
        if (!TransimsLinkStatOutput->Open(Filename, "w"))
        {
            Error("cannot open TRANSIMS link stat file %s", Filename);
            return FALSE;
        }
        TransimsLinkStatOutput->Write(TRANSIMS_LINK_STAT_FILE_HEADER);
        Print("TRANSIMS link stat file %s opened", Filename);
    }
    return TRUE;
}

// ===================================================================================

bool
TRouter::ReadRoute(long DepartureTimeStep)

{
    TVehicle      *NewVehicle;
    TRouteplan    *Routeplan = 0;
    TNode         *Node = 0;
    TAccessory    *Accessory;
    TTransferInfo  TransferInfo;
    bool           VehicleAdded = FALSE;

    //  >>> DO NOT SEPERATE <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
#   define READ_ROUTE_LONG 2 + SIZE_OF_PARALLEL_ID_IN_LONG
    long           RouteplanFlags;
    TParallelID    SourceID;
    //  >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

    TransferInfo.TransferMode = TM_AsciiData;
    if (!(RouteplanParser->Decode(RouteplanFlags) AND
          RouteplanParser->Decode(SourceID)))
        return FALSE;

    Print(DM_ROUTE, "ReadRoute %ld Flags:%ld, ID:%lld", DepartureTimeStep, RouteplanFlags, SourceID);

    RouteplanFlags |= RP_FLAG_SHORTEN_ON_ENCODE;
    if ((RouteplanFlags & RP_FLAG_SOURCE_IS_ACCESSORY) AND
        (Accessory = TAccessory::GetPtr(SourceID)))
    {
        Routeplan = new TRouteplan(DepartureTimeStep, RouteplanFlags, SourceID,
                                   RouteplanParser, &TransferInfo);
        NewVehicle = new TVehicle(Routeplan, RouteplanParser, &TransferInfo);
        if (!Accessory->AddVehicle(NewVehicle))
            return FALSE;
        VehicleAdded = TRUE;
    }
    else
    {
        TInfoID        InfoID;
        TIntersection *Intersection = 0;

        InfoID = TSimpleIntersection::ClassInfo->GetInfoID();
        if ((RouteplanFlags & RP_FLAG_SOURCE_IS_ACCESSORY) == 0 AND 
            (Node = TNode::GetPtr(SourceID)) AND
            (Intersection = (TIntersection *) Node->GetInfo(InfoID)))
        {
            Routeplan = new TRouteplan(DepartureTimeStep, RouteplanFlags, SourceID, 
                                       RouteplanParser, &TransferInfo);         
            NewVehicle = new TVehicle(Routeplan, RouteplanParser, &TransferInfo);
            if (!Intersection->AddVehicle(NewVehicle))
                return FALSE;
            VehicleAdded = TRUE;
        }
    }
        
    // goto to next routeplan header

    if (BinaryRouteplan)
    {
        if (!VehicleAdded)
            if (!(SkipRouteplan->Skip(RouteplanParser) AND
                  SkipVehicle->Decode(RouteplanParser, &TransferInfo)))
            {
                Error("ReadRoute: cannot skip to next route-plan");
                return FALSE;
            }
    }
    else
        if (!RouteplanParser->FindToken(TOKEN_ROUTE))
        {
            Error("ReadRoute: cannot find ROUTE token");
            return FALSE;
        }
    return TRUE;
}

// =================================================================================== 

bool
TRouter::Reroute(long CurrentTimeStep, TParallelID CurrentEdgeID,
                 double RelativeLocation, TVehicle *Vehicle,
                 int LastValidStep, long EstimatedDeletionTimeStep)

{
    TRouteplan     *Routeplan;
    TNetAccessInfo  AccessInfos[2];
    TNetAccessInfo *DestAccessInfo;
    TRouteStep      RouteSteps[MAX_ROUTE_LENGTH];
    int             NrOfRouteSteps;
    int             NrOfDestAccessInfos;
    long            RerouteArrivalTime, DetourArrivalTime = INVALID;
    long            Improvement;
    double          RelativeImprovement;
    bool            ComputeDetour;
    int             RerouteType, RerouteResult, EstimateSteps;
    long            RemainingPlanLength;
    long            ArrivalTimeStep;

    Routeplan = Vehicle->GetRouteplan();
    ComputeDetour = (LastValidStep < Routeplan->GetNrOfWholeSteps() - 1);
    EstimateSteps = LastValidStep - Routeplan->GetCurrentStep() + 1;
    RemainingPlanLength = Routeplan->GetRemainingPlanLength();
    ArrivalTimeStep = Routeplan->GetArrivalTimeStep();
    WriteReroutingLog("%ld %ld %.2f", CurrentTimeStep,
                      Routeplan->GetRouteID(), Routeplan->GetEstimatedRelativeDelay());
    if (RemainingPlanLength < REROUTE_MIN_REMAINING_LENGTH)
    {
        WriteReroutingLog(" %d %d\n", REROUTE_TYPE_TOO_CLOSE_TO_DEST, RemainingPlanLength);
        return TRUE;
    }
    if ((AccessInfos[0].Edge = TEdge::GetPtr(CurrentEdgeID)) == 0)
    {
        Error("Reroute: cannot retrieve current edge %ld\n", CurrentEdgeID);
        return FALSE;
    }    
    if ((AccessInfos[0].Node = AccessInfos[0].Edge->GetToNode()) == 0)
    {
        Error("Reroute: cannot retrieve ToNode of current edge %s",
              (char *) AccessInfos[0].Edge->GetInstanceName());
        return FALSE;
    }
    AccessInfos[0].FractionOnEdge = 1.0 - RelativeLocation;
    AccessInfos[0].Time = CurrentTimeStep;
    if (!LocalSP->BuildTree(AccessInfos, 1))
    {
        WriteReroutingLog(" %d %d\n", REROUTE_ERROR_BUILD_TREE, (int) LocalSP->GetError());
        return TRUE;
    }
    NrOfDestAccessInfos = 0;
    if (ComputeDetour)
    {
        // LastValidStep is the last step along the current route for which
        // we have found a valid edge weight. We will compute a shortest path
        // to the FromNode of that link.

        TEdge       *Edge;      
        TNode       *DetourNode;
        TParallelID  LastValidEdgeID;

        RerouteType = REROUTE_TYPE_DETOUR;
        LastValidEdgeID = Routeplan->GetStep(LastValidStep).ID;
        if ((Edge = TEdge::GetPtr(LastValidEdgeID)) == 0)
        {
            Error("Reroute: cannot retrieve last valid edge %ld", LastValidEdgeID);
            return FALSE;
        }
        if ((DetourNode = Edge->GetToNode()) == 0)
        {
            Error("Reroute: cannot retrieve detour node for edge %s",
                  (char *) Edge->GetObjectName());
            return FALSE;
        }
        AccessInfos[NrOfDestAccessInfos].Edge = 0;
        AccessInfos[NrOfDestAccessInfos].Node = DetourNode;
        NrOfDestAccessInfos++;
    }
    else
    {
        // All edge weights to the destination were available. We have to differentiate
        // between the source being a node or an accessory.

        RerouteType = REROUTE_TYPE_REROUTE;
        if (Routeplan->DestinationIsAccessory())
        {
            TParallelID  AccID;
            TEdge       *DestEdge;
            
            // Here we cheat a little bit: since it's pretty time-consuming to 
            // load data about the accessory in question we simply assume that
            // the accessory ID matches the edge ID and that the accessory is
            // located at the center of the edge.
            
            AccID = Routeplan->GetDestID();
            if ((DestEdge = TEdge::GetPtr(AccID)))
            {
                TNode *FromNode;
                
                if ((FromNode = DestEdge->GetFromNode()) == 0)
                {                   
                    Error("Reroute: cannot retrieve FromNode for accessory %ld\n", AccID);
                    return FALSE;
                }
                AccessInfos[NrOfDestAccessInfos].Edge = DestEdge;
                AccessInfos[NrOfDestAccessInfos].Node = FromNode;
                AccessInfos[NrOfDestAccessInfos].FractionOnEdge = 0.5;
                NrOfDestAccessInfos++;
            }
            
        /*

        // We would like to enable the driver to approach the accessory from
        // 'the opposite direction' than the one that was encoded in his original
        // route-plan. So we flip the least-significant bit of the accessory/edge ID
        // which should give us the opposite direction.

        AccID ^= 1;
        if ((DestEdge = TEdge::GetPtr(AccID)))
        {
            TNode *FromNode;

            if ((FromNode = DestEdge->GetFromNode()) == 0)
            {
                WriteReroutingLog(" ERROR: cannot retrieve FromNode of last edge %ld, AccID\n");
                return TRUE;
            }
            AccessInfos[NrOfDestAccessInfos].Edge = DestEdge;
            AccessInfos[NrOfDestAccessInfos].Node = FromNode;
            AccessInfos[NrOfDestAccessInfos].FractionOnEdge = 0.5;
            NrOfDestAccessInfos++;
        }
        */
        }
        else // the route terminates on a node
        {

            TNode *DestNode;

            if ((DestNode = TNode::GetPtr(Routeplan->GetDestID())) == 0)
            {
                WriteReroutingLog("Reroute: cannot retrieve destination node %ld",
                                  Routeplan->GetDestID());
                return FALSE;
            }
            else
            {
                AccessInfos[NrOfDestAccessInfos].Edge = 0;
                AccessInfos[NrOfDestAccessInfos].Node = DestNode;
                NrOfDestAccessInfos++;
            }
        }
    }

    if (!LocalSP->GetPath(AccessInfos, NrOfDestAccessInfos, 
                          RouteSteps, MAX_ROUTE_LENGTH, NrOfRouteSteps, DestAccessInfo))
    {
        WriteReroutingLog(" %d %d %d\n", REROUTE_ERROR_GET_PATH, (int) LocalSP->GetError(),
                          RerouteType);
        return TRUE;
    }
    if (ComputeDetour)
    {
        long RemainingTourTime;

        RemainingTourTime = Routeplan->GetArrivalTimeStep() -
            Routeplan->GetStep(LastValidStep + 1).Estimate;
        DetourArrivalTime = DestAccessInfo->Time;
        RerouteArrivalTime = DetourArrivalTime + RemainingTourTime;
    }
    else    
        RerouteArrivalTime = DestAccessInfo->Time;

    //    !!!!!!!!! make this configurable!!!
    //    Improvement = Routeplan->GetArrivalTimeStep() - RerouteArrivalTime;
    Improvement = EstimatedDeletionTimeStep - RerouteArrivalTime;
    //    !!!!!!!!!

    RelativeImprovement = 1.0 * Improvement / Routeplan->GetEstimatedTravelTime();
    if (RelativeImprovement > ReroutingMinRelativeImprovement)
        if (Routeplan->ContinuesWith(RouteSteps, NrOfRouteSteps))
            RerouteResult = REROUTE_RESULT_SAME_ROUTE;
        else
        {
            RerouteResult = REROUTE_RESULT_NEW_ROUTE;
            if (ComputeDetour)
            {
                if (!Routeplan->Detour(CurrentTimeStep, LastValidStep, RouteSteps, 
                                       NrOfRouteSteps, DetourArrivalTime))
                {
                    Error("Reroute: cannot replace route by detour");
                    return FALSE;
                }
            }
            else
                if (!Routeplan->Reroute(CurrentTimeStep, RouteSteps, NrOfRouteSteps, 
                                        RerouteArrivalTime))
                {
                    Error("Reroute: cannot replace route");
                    return FALSE;
                }
        }
    else
        RerouteResult = REROUTE_RESULT_KEEP_ROUTE;
    WriteReroutingLog(" %d %d %d %d %ld %ld %ld %.2f %d\n", RerouteType,
                      RemainingPlanLength, EstimateSteps, NrOfRouteSteps,
                      ArrivalTimeStep, EstimatedDeletionTimeStep,
                      RerouteArrivalTime, RelativeImprovement, RerouteResult);
    return TRUE;
}

// =================================================================================== 

void
TRouter::Reset(void)

{
    RoutesProcessed = 0;
    RoutesValid = 0;
    RoutesCorrupted = 0;
    RoutesOutsideRegion = 0;
    RoutesStartingOutsideRegion = 0;
    RoutesEndingOutsideRegion = 0;
    RoutesOutsideTimeWindow = 0;
    RoutesWritten = 0;
    SumRouteLength = 0;
    MinRouteLength = 99999;
    MaxRouteLength = 0;    
    ConversionLogEntries = 0;
}

// =================================================================================== 

bool
TRouter::Run(const TTimeStep &theStartTimeStep)

{    
    StartTimeStep = theStartTimeStep;
    if (!OpenRouteplanForReading())
    {
        Error("Run: cannot open routeplan for reading");
        return FALSE;
    }
    if (Parallel->IsMaster())
    {
        if (!OpenCompletedRoutesOutput())
        {
            Error("Run: cannot open completed routes file");
            return FALSE;
        }
        if (!InitRoutesByOrigin())
        {
            Error("Run: cannot init routes by origin statistics");
            return FALSE;
        }
        if (!OpenTransimsLinkStatOutput())
        {
            Error("Run: cannot open TRANSIMS link stat output");
            return FALSE;
        }
    }
    if (!ActivateStatistics())
    {
        Error("Run: cannot activate statistics");
        return FALSE;
    }
    if (!StatManager->RegisterSamplingHandler(TRoutingStat::ClassInfo->GetID(), this))
    {
        Error("Run: cannot register myself as sampling handler");
        return FALSE;
    }
    if (!ActivateLocalRerouting())
    {
        Error("cannot activate local re-rerouting");
        return FALSE;
    }
    return TRUE;
}

// =================================================================================== 

bool
TRouter::SendExecutedRouteplans(void)

{
    if (ExecutedRouteplans.GetNrOfElements() > 0)
    {
        TBaseMessage *Message = 0;

        if (!(Communication->GetNewMessage(Message, DEST_PARENT, MTAG_ROUTER,
                                           MTYPE_ROUTER) AND
              Message->Event(MEVENT_ROUTER_COMPLETED_ROUTE) AND
              ObjectManager->SendObjectList((TParallelObjectList *) &ExecutedRouteplans, 
                                            Message) AND
              Communication->Commit(Message)))
        {
            Error("HandleExecutedRouteplan: cannot send completed route");
            return FALSE;
        }
        ExecutedRouteplans.DeleteAll(TRUE);
    }
    return TRUE;
}

// =================================================================================== 

bool
TRouter::SortRouteplans(void) const

{
    TRouteInfo *RouteInfos;
    long        i;
    char        RouteBuffer[ROUTE_BUFFER_SIZE];

    if ((RouteInfos = (TRouteInfo *) malloc(RoutesWritten * sizeof(TRouteInfo))) == 0)
    {
        Error("SortRouteplans: cannot allocate array for %ld route infos",
              RoutesWritten);
        return FALSE;
    }
    Print("reading route infos...");
    fseek(OffsetFile, 0, SEEK_SET);
    fread(RouteInfos, sizeof(TRouteInfo), RoutesWritten, OffsetFile);
    Print("sorting routes...");
    qsort(RouteInfos, RoutesWritten, sizeof(TRouteInfo), 
          (int(*)(const void*, const void*)) CompareRouteInfos);
    Print("writing sorted routes...");
    for (i = 0; i < RoutesWritten; i++)
    {
        fseek(RouteplanTempOutput->GetFile(), RouteInfos[i].Offset, SEEK_SET);
        fread(RouteBuffer, RouteInfos[i].Length, 1, RouteplanTempOutput->GetFile());
        fwrite(RouteBuffer, RouteInfos[i].Length, 1, RouteplanOutput->GetFile());
    }    
    return TRUE;
}

/* =================================================================================== */

bool
TRouter::StoreGridStats(TClassInfo *StatClassInfo, TGridStatList *StatList)

{
    TInfoID   InfoID;

    InfoID = StatClassInfo->GetInfoID();
    if (MaxGridStatEntries != 0)
    {
        TGridStatEnum CStat;
        TInfoMask     LocalInfoMask;
    
        LocalInfoMask = StatClassInfo->GetInfoMask();
        CStat.WalksForwardThrough(StatList);
        while (CStat())
        {       
            TEdge           *Edge;
            TGridStatLookup *Lookup;

            if ((Edge = TEdge::LoadInstance(CStat->GetParallelID(), 
                                            LocalInfoMask, TRUE)) == 0)
            {
                Error("StoreGridStats: cannot load edge %ld", CStat->GetParallelID());
                return FALSE;
            }
            if ((Lookup = (TGridStatLookup *) Edge->GetInfo(InfoID, TRUE, TRUE)) == 0)
            {
                Error("StoreGridStats: cannot find grid stat time lookup for %s",
                      (char *) Edge->GetObjectName());
                return FALSE;
            }
            /*
            if (CStat->GetNrOfVehicles() > 0)
                Print("%s for %s", (char *) CStat->GetObjectName(),
                      (char *) StatClassInfo->GetInstanceName());
                      */
            Lookup->Put(CStat);
            CStat++;
        }
    }
    StatList->DeleteAll();
    ObjectManager->Action(StatClassInfo, 0, ActionModified);
    return TRUE;
}

/* =================================================================================== */

bool
TRouter::UpdateGridStatDestList(void)

{
    TEnumerator< TExtendedCPNInfo > CInfo (TExtendedCPNInfo::ClassInfo->GetTable());
    TExtendedCPNInfo *LocalInfo;
    TMapStat         *MapStat;
    TInfoID           InfoID;
    TRectangle        LocalMapRectangle;

    if ((LocalInfo = Network->GetExtendedCPNInfo(Parallel->GetLocalCPNID())) == 0)
    {
        Error("UpdateGridStatDestList: cannot get local CPNInfo");
        return FALSE;
    }
    InfoID = TMapStat::ClassInfo->GetInfoID();
    if ((MapStat = (TMapStat *) LocalInfo->GetInfo(InfoID)) == 0)
    {
        Error("UpdateGridStatDestList: cannot get local map stat");
        return FALSE;   
    }
    LocalMapRectangle = MapStat->GetMapRectangle();
    LocalMapRectangle += LocalGridStatOverlap;

    //    Print("overlap %f local %s enlarged %s", LocalGridStatOverlap,
    //  (char *) MapStat->GetMapRectangle().GetInstanceName(),
    //  (char *)  LocalMapRectangle.GetInstanceName());
    
    LocalGridDestList.DeleteAll();
    CInfo.Enumerate();
    while (CInfo())
    {
        if ((MapStat = (TMapStat *) CInfo->GetInfo(InfoID)) == 0)
        {
            Error("UpdateGridStatDestList: cannot get local map stat");
            return FALSE;       
        }
        if (LocalMapRectangle.Overlaps(MapStat->GetMapRectangle()))
            LocalGridDestList.Append(CInfo->CPNInfo);
        CInfo++;
    }
    return TRUE;
}

/* =================================================================================== */

bool
TRouter::WriteLog(const char *Format, ...)
    
{
    va_list argptr;
    char    line[2048];

    if (!OpenConversionLogFile())
        return FALSE;
    va_start(argptr, Format);
    vsprintf(line, Format, argptr);
    fprintf(ConversionLogOutput, "%s\n", line);     
    va_end(argptr);
    ConversionLogEntries++;
    return TRUE;
}

/* =================================================================================== */

bool
TRouter::WriteReroutingLog(const char *Format, ...)
    
{
    va_list      argptr;
    char         line[2048];
    static long  Counter = 0;

    if (ReroutingLog == 0)
        return TRUE;
    va_start(argptr, Format);
    vsprintf(line, Format, argptr);
    ReroutingLog->Write(line);
    va_end(argptr);
    if (Counter++ & 15 == 0)
        ReroutingLog->Flush();
    return TRUE;
}

/* =================================================================================== */

bool
TRouter::WriteRoute(TVehicle *Vehicle)

{
    TStream       *Output;
    TRouteInfo     RouteInfo;
    TRouteplan    *Routeplan;
    TTransferInfo  TransferInfo;

    TransferInfo.TransferMode = TM_AsciiData;
    if ((Routeplan = Vehicle->GetRouteplan()) == 0)
    {
        Error("WriteRoute: no routeplan");
        return FALSE;
    }
    if (SortByDepartureTime)
    {
        RouteInfo.Offset = ftell(RouteplanTempOutput->GetFile());
        RouteInfo.DepartureTime = Routeplan->GetDepartureTimeStep();
        Output = RouteplanTempOutput;
    }
    else
        Output = RouteplanOutput;
    if (!(Vehicle->GetRouteplan()->Encode(Output, &TransferInfo) AND
          Vehicle->Encode(Output, &TransferInfo)))
    {
        Error("WriteRoute: cannot write routeplan and vehicle data");
        return FALSE;
    }
    if (SortByDepartureTime)
    {
        RouteInfo.Length = ftell(RouteplanTempOutput->GetFile()) - RouteInfo.Offset;
        RouteInfo.SourceID = Routeplan->GetSourceID();
        fwrite(&RouteInfo, sizeof(TRouteInfo), 1, OffsetFile);
    }
    RoutesWritten++;
    return TRUE;
}

/* =================================================================================== */

bool
TRouter::WriteRouteOriginStat(TRouteplan *Routeplan)

{
    TLocation SourceLocation;
    TLocation DestLocation;
    double    Distance;
    double    Angle;
    int       Sector;

    if (Routeplan->SourceIsAccessory())
        return TRUE;
    if (!GetLocation(Routeplan->GetDestID(), Routeplan->DestinationIsAccessory(), DestLocation))
    {
        Warning("WriteRouteOriginStat: cannot retrieve dest location for %ld",
                Routeplan->GetDestID());
        return TRUE;
    }
    if (!RouteDestRectangle.Contains(DestLocation))
        return TRUE;
    if (!GetLocation(Routeplan->GetSourceID(), FALSE, SourceLocation))
    {
        Warning("WriteRouteOriginStat: cannot retrieve source location for %ld",
                Routeplan->GetSourceID());
        return TRUE;
    }
    Distance = LengthInMeter(SourceLocation.DistanceTo(DestLocation));
    Angle = RouteDestLocation.AngleTo(SourceLocation) / (2.0 * PI);
    Sector = (int) (Angle * RouteOriginSectors);
    RoutesByOriginOutput->Write("%ld %ld %ld %.1f %.1f %.1f %.1f %.1f %.1f %f %d\n", 
                                Routeplan->GetInsertionTimeStep(),
                                Routeplan->GetDeletionTimeStep(),
                                Routeplan->GetActualTravelTime(),
                                SourceLocation.x(), SourceLocation.y(),
                                DestLocation.x(), DestLocation.y(),
                                Distance,
                                Distance / Routeplan->GetActualTravelTime(),
                                Angle,
                                Sector);
    return TRUE;
}

/* =================================================================================== */

bool
TRouter::WriteStat(const char *Format, ...)
    
{
    va_list argptr;
    char    line[2048];

    if (!OpenConversionStatFile())
        return FALSE;
    va_start(argptr, Format);
    vsprintf(line, Format, argptr);
    fprintf(ConversionStatOutput, "%s\n", line);            
    va_end(argptr);
    return TRUE;
}

// =================================================================================== 

bool
TRouter::WriteStatistics(void)

{
    WriteStat("%c--routing statistics (TRouter)---------", COMMENT_CHAR);
    if (RoutesProcessed)
    {
        WriteStat("%croutes processed                      = %ld", COMMENT_CHAR, 
                  RoutesProcessed);
        WriteStat("%croutes valid                          = %ld", COMMENT_CHAR, 
                  RoutesValid);
        WriteStat("%croutes outside time window            = %ld", COMMENT_CHAR, 
                  RoutesOutsideTimeWindow);
        WriteStat("%croutes outside region                 = %ld", COMMENT_CHAR, 
                  RoutesOutsideRegion);
        if (RoutesValid > 0)
        {
            WriteStat("%croutes starting inside/outside region = %ld / %ld", COMMENT_CHAR, 
                      RoutesValid - RoutesStartingOutsideRegion, 
                      RoutesStartingOutsideRegion);
            WriteStat("%croutes ending inside/outside region   = %ld / %ld ", COMMENT_CHAR, 
                      RoutesValid - RoutesEndingOutsideRegion,
                      RoutesEndingOutsideRegion);
            WriteStat("%cshortest route                        = %ld", COMMENT_CHAR, 
                      MinRouteLength);
            WriteStat("%clongest route                         = %ld", COMMENT_CHAR, 
                      MaxRouteLength);
            WriteStat("%caverage route length                  = %f", COMMENT_CHAR, 
                      (double) SumRouteLength / RoutesValid);
        }           
    }
    else
        WriteStat("%cno routes processed", COMMENT_CHAR);
    return TRUE;
}

// =================================================================================== 

bool
TRouter::WriteTransimsLinkStat(TGridStatList *StatList)

{
    if (TransimsLinkStatOutput)
    {
        TGridStatEnum CStat;
        
        CStat.WalksForwardThrough(StatList);
        while (CStat())
        {       
            if ((long) CStat->GetTimeStep() > 0)
                CStat->WriteTransimsLinkStat(TransimsLinkStatOutput);
            CStat++;
        }
    }
    return TRUE;
}

// =================================================================================== 
#undef ParentClass

