/*      file     : TrafficData/TSimpleIntersection.C
        project  : PAMINA III
        author   : Marcus Rickert
        RCS      : $Id: TSimpleIntersection.C,v 1.9 2005/03/27 20:39:16 mr Exp $

        Copyright (C) 2005 Marcus Rickert

        http://www.the-rickerts.de/mr/PAMINA.html
*/

// ***INCLUDES************************************************************************ 

#include <TString.h>
#include <TBaseStream.h>
#include <TRandom.h>

#include <TIncidenceInfo.h>
#include <TEdge.h>

#include "CAConfig.h"
#include "TEdgeInfo.h"
#include "TRouteplan.h"
#include "TSink.h"
#include "TMultilaneGrid.h"
#include "TSimpleIntersection.h"

// ***DEFINES*************************************************************************

#define EDGE_FREEWAY_TYPE 1
#define EDGE_RAMP_TYPE    6

// ***GLOBAL VARIABLES**************************************************************** 

TClassInfo  *TSimpleIntersection::ClassInfo = 0;
int          TSimpleIntersection::ScanRange = 5;
int          TSimpleIntersection::TVelocity = 1;

// ***METHODS************************************************************************* 

#ifdef NO_INLINE
#  include <TSimpleIntersection.i.C>
#endif

// =================================================================================== 
#define ParentClass TIntersection
// ===================================================================================


TSimpleIntersection::TSimpleIntersection(TParallelID theParallelID) :

ParentClass(ClassInfo, theParallelID)

{
    DestGridNrs = 0;
    Random = new TRandom;
}

// ===================================================================================

TSimpleIntersection::~TSimpleIntersection(void)

{
    DeleteObjectArray(DestGridNrs);
    DeleteObject(Random);
}

// ===================================================================================

bool
TSimpleIntersection::Activate(void)

{
    if (!ParentClass::Activate())
        return FALSE;
    if (!DetectThroughways())
    {
        Error("Activate: while detecting throughways");
        return FALSE;
    }
    Block();
    return TRUE;
}

// ===================================================================================

void
TSimpleIntersection::Block(void)

{
    int             CGrid;
    int             CLane;
    TMultilaneGrid *Grid;
    
    for (CGrid = 0 ; CGrid < IncidenceInfo->GetNrOfIncomingEdges(); CGrid++)
    {
        Grid = IncomingGrids[CGrid];
        for (CLane = 0; CLane < Grid->GetNrOfLanes(); CLane++)
        {
            // get a 'pointer' to the last 'SimpleIntersectionCheckRange'
            // sites on the current lane

            long LastSite = Grid->GetLastSite();
            Grid->SAccess(CLane, LastSite) |= SITE_BLOCKED;
        }                   
    }
}

// =================================================================================== 

bool
TSimpleIntersection::DetectThroughways(void)

{
    int     CInEdge;
    int     COutEdge;
    TInfoID InfoID;
    int     NrOfThroughwayEdges;

    NrOfThroughwayEdges = 0;
    DestGridNrs = new int [IncidenceInfo->GetNrOfIncomingEdges()];
    InfoID = TEdgeInfo::ClassInfo->GetInfoID();
    for (CInEdge = 0; CInEdge < IncidenceInfo->GetNrOfIncomingEdges(); CInEdge++)
    {
        TEdge     *InEdge;
        TEdgeInfo *InEdgeInfo;

        InEdge = IncidenceInfo->GetIncomingEdge(CInEdge);
        if ((InEdgeInfo = (TEdgeInfo *) InEdge->GetInfo(InfoID)) == 0)
        {
            Error("DetectThroughways: cannot read edge info for %d",
                  (char *) InEdge->GetObjectName());
            return FALSE;
        }
        if (InEdgeInfo->Class == EDGE_THROUGHWAY_CLASS)
        {        
            TEdge     *OutEdge = 0;
            TEdgeInfo *OutEdgeInfo = 0;
            bool       Found;

            NrOfThroughwayEdges++;
            if (NrOfThroughwayEdges > 2)
            {
                Error("number of incoming throughway edges > 2");
                return FALSE;
            }
            COutEdge = 0; 
            Found = FALSE;
            while (COutEdge < IncidenceInfo->GetNrOfOutgoingEdges() AND !Found)
            {           
                OutEdge = IncidenceInfo->GetOutgoingEdge(COutEdge);
                if ((OutEdgeInfo = (TEdgeInfo *) OutEdge->GetInfo(InfoID)) == 0)
                {
                    Error("DetectThroughways: cannot read edge info for %d",
                          (char *) OutEdge->GetObjectName());
                    return FALSE;
                }
                if (OutEdgeInfo->Class == EDGE_THROUGHWAY_CLASS AND
                    OutEdge->GetToNodeID() != InEdge->GetFromNodeID())
                    Found = TRUE;
                else
                    COutEdge++;
            }
            if (Found)
            {
                //                Print("detected throughway from %s (class:%d) to %s (class:%d) ",
                //                      (char *) InEdge->GetObjectName(), InEdgeInfo->Class,
                //                      (char *) OutEdge->GetObjectName(), OutEdgeInfo->Class);
                DestGridNrs[CInEdge] = COutEdge;
            }
            else
                DestGridNrs[CInEdge] = INVALID;
        }
        else
            DestGridNrs[CInEdge] = INVALID;
    }
    return TRUE;
}

// ===================================================================================

bool
TSimpleIntersection::ExecuteSubTimeStep(const TTimeStep &TimeStep)

{

    int             CGrid, CCorGrid;
    TMultilaneGrid *Grid = 0;
    int             Shift;

    if (!ParentClass::ExecuteSubTimeStep(TimeStep))
        return FALSE;
    Shift = Random->Int(IncidenceInfo->GetNrOfIncomingEdges());
    // walk through all incident edges...
    for (CGrid = 0; CGrid < IncidenceInfo->GetNrOfIncomingEdges(); CGrid++)
    {
        CCorGrid = (Shift + CGrid) % IncidenceInfo->GetNrOfIncomingEdges();
        if (IsIncomingSegmentActive(CCorGrid))
        {
            Grid = IncomingGrids[CCorGrid];
            long LastSite = Grid->GetLastSite();
            if (Grid->VehicleRangeIsActive(LastSite - ScanRange, ScanRange))
                if (!ScanSourceGrid(TimeStep, CCorGrid))
                    return FALSE;
        }
    }
    for (CGrid = 0; CGrid < IncidenceInfo->GetNrOfIncomingEdges(); CGrid++)
    {
        if (DestGridNrs[CGrid] != INVALID)
            
            // the current grid has a simple connection to an outgoing grid:

            if (!OutgoingGrids[DestGridNrs[CGrid]]->ExchangeBoundaries(IncomingGrids[CGrid]))
            {
                Error("ExecuteSubTimeStep: cannot exchange boundaries from %s to %s",
                      (char *) IncomingGrids[CGrid]->GetObjectName(),
                      (char *) OutgoingGrids[DestGridNrs[CGrid]]->GetObjectName());
                return FALSE;
            }
    }
    return TRUE;
}

// ===================================================================================

TString
TSimpleIntersection::GetClassName(void) const

{
    return "TSimpleIntersection";
}

// ===================================================================================

bool
TSimpleIntersection::MoveToDestinationSegment(TVehicle *Vehicle, int SourceGridNr,
                                              long Lane, long Offset, long Vmax)

{
    TMultilaneGrid *DestGrid = 0, *SourceGrid;
    TRouteplan     *Routeplan;
    TParallelID     DestEdgeID;
    long            DestGridNr;
    long            DestLane;
    long            CLane, CorDestLane = 0;
    bool            Found;
    long            i;


    // Print("MoveToDestSegment: lane:%ld offset:%ld", Lane, Offset);

    SourceGrid = IncomingGrids[SourceGridNr];
    if ((Routeplan = Vehicle->GetRouteplan()))
    {
        if (Routeplan->DestinationIsAccessory() AND
            Routeplan->GetRemainingPlanLength() == 0)
        {
            //Warning("ignoring route %s", (char *) Routeplan->GetInstanceName());
            //      SourceGrid->RemoveVehicle(Lane, Offset, FALSE);
            //      VehicleFailed(Vehicle);
            return TRUE;
        }
        DestEdgeID = Routeplan->GetCurrentID();

        if ((DestGridNr = GetOutgoingGridNrByID(DestEdgeID)) == INVALID)
        {
            static char Format[128] = { 0 };

            if (0 == Format[0]) {
                sprintf(Format, "cannot find edge ID %s for route %%s", PARALLEL_ID_PRINTF_FORMAT);
            }
            Warning(Format, DestEdgeID, (char *) Routeplan->GetInstanceName());
            SourceGrid->RemoveVehicle(Lane, Offset, FALSE);
            VehicleFailed(Vehicle);
            return TRUE;
        }

        //      Print("DestGridNr:%ld SourceGridNr:%ld DestGridNrs[SourceGridNr]:%ld", 
        //            DestGridNr, SourceGridNr, DestGridNrs[SourceGridNr]);

        //      Print("MoveToSegment %s", (char *) Routeplan->GetInstanceName());

        if (DestGridNr == DestGridNrs[SourceGridNr])
        {
            //      Print("vehicle on highway v=%ld",
            //            SITE_GET_VELOCITY(SourceGrid->SAccess(Lane, Offset)));


            return TRUE;
        }
    }
    else
    {
        Warning("no routeplan!");
        SourceGrid->RemoveVehicle(Lane, Offset, FALSE);
        VehicleFailed(Vehicle);
        return TRUE;
    }
    DestGrid = OutgoingGrids[DestGridNr]; 
    // choose a lane of the destination grid, preferably the same
    // one as on the previous grid
    DestLane = min(Lane, DestGrid->GetNrOfLanes() - 1);
    Found = FALSE;
    CLane = 0; 
    while (CLane > - DestGrid->GetNrOfLanes() AND !Found)
    {
        CorDestLane = (DestLane + CLane + DestGrid->GetNrOfLanes()) %
                      DestGrid->GetNrOfLanes();
        i = 0;
        while (i < TVelocity AND DestGrid->VAccess(CorDestLane, i) == 0)
            i++;
        if (i == TVelocity)
            Found = TRUE;
        else
            CLane--;
    }
    if (Found) // <- contains averaging over phases (if active)
    {
        // ... if there is room for it!
        
        SourceGrid->RemoveVehicle(Lane, Offset, FALSE);

        // insert vehicle into the VERY FIRST site in destination lane

        Vehicle->SetSitePattern(DestGrid->SAccess(CorDestLane, 0));


        // increase velicity to at least TVelocity
        if (Vehicle->GetCAVelocity() < TVelocity)
        {
            Vehicle->SetCurrentVelocityInCAUnits(TVelocity);
            // Print("velocity set to %ld", NodeInfo->SimpleIntersectionTVelocity);
        }
        DestGrid->SetVehicle(CorDestLane, 0, Vehicle, Vehicle->GetSitePattern());
        Turn(SourceGridNr, DestGridNr);
        if (Routeplan)
            Routeplan->Step();
    }
    else
    {
        // slow down vehicles that cannot be inserted into destination grid

        SITE_SET_V_MAX(SourceGrid->SAccess(Lane, Offset), Vmax);


    }
    return TRUE;
}

// ===================================================================================

bool
TSimpleIntersection::PrepareSubTimeStep(const TTimeStep &)

{

    int CGrid;

    //    if (GetParallelID() == 7451)
    //  Print("PrepareSubStimeStep");

    // walk through all incident edges...
    for (CGrid = 0; CGrid < IncidenceInfo->GetNrOfIncomingEdges(); CGrid++)
        if (DestGridNrs[CGrid] != INVALID)
        {
            
            // the current grid has a simple connection to an outgoing grid:

            if (GetParallelID() == 7451)
                Print("simple connection");
            
            if (!OutgoingGrids[DestGridNrs[CGrid]]->AbsorbVehicles(IncomingGrids[CGrid]))
            {
                Error("ExecuteSubTimeStep: cannot move vehicles from %s to %s",
                      (char *) IncomingGrids[CGrid]->GetObjectName(),
                      (char *) OutgoingGrids[DestGridNrs[CGrid]]->GetObjectName());
                return FALSE;
            }
        }
    return TRUE;
}

// ===================================================================================

bool
TSimpleIntersection::ScanSourceGrid(const TTimeStep &TimeStep, int SourceGridNr)

{
    int                      CLane, CorLane, StartLane;
    int                      i;
    int                      VehicleFound[MAX_LANES], VehiclesFound;
    long                     ScanOffset;
    TMultilaneGrid          *SourceGrid;
    TVehicle                *Vehicle;

    SourceGrid = IncomingGrids[SourceGridNr];
    ScanOffset = SourceGrid->GetLastSite() - ScanRange;

    // Print("ScanSourceGrid: grid:%ld offset:%ld", SourceGridNr, ScanOffset);

    for (CLane = 0; CLane < SourceGrid->GetNrOfLanes(); CLane++)
        VehicleFound[CLane] = FALSE;
    i = ScanRange - 1;
    VehiclesFound = 0;
    StartLane = Random->Int(SourceGrid->GetNrOfLanes());
    while (i >= 0 AND VehiclesFound < SourceGrid->GetNrOfLanes())
    {
        for (CLane = 0; CLane < SourceGrid->GetNrOfLanes(); CLane++)
        {
            // make sure we treat all incoming lanes in a robin-round fashion
            
            CorLane = (CLane + StartLane) % SourceGrid->GetNrOfLanes();
            
            if (!VehicleFound[CorLane])
            {
                
                // check for vehicles in all sites of that range...
                
                if ((Vehicle = SourceGrid->VAccess(CorLane, ScanOffset + i)))
                {
                    // ... and eventually find a vehicle...

                    if (Vehicle AND IsIncomingVehicleActive(SourceGridNr))
                    {

                        TRouteplan *Routeplan;

                        if ((Routeplan = Vehicle->GetRouteplan()) AND
                            Routeplan->IsDestinationNode(GetParallelID()))
                        {
                            SourceGrid->RemoveVehicle(CorLane, ScanOffset + i, FALSE);
                            Sink->DestroyVehicle(TimeStep, Vehicle);
                        }
                        else
                        {                           


                            if (!MoveToDestinationSegment(Vehicle, SourceGridNr, CorLane,
                                                          ScanOffset + i, 
                                                          CA_DEFAULT_V_MAX - 1 - i))
                                return FALSE;
                            /// TODO: remove the check for VehicleFound?

                            //                      VehicleFound[CorLane] = TRUE;
                            //                      VehiclesFound++;
                        }
                    }
                }
            }
        }
        i--;
    }
    return TRUE;
}

// ===================================================================================
#undef ParentClass
