/*      file     : TrafficData/TSource.C
        project  : PAMINA III
        author   : Marcus Rickert
        RCS      : $Id: TSource.C,v 1.6 2005/03/25 12:28:52 mr Exp $

        Copyright (C) 2005 Marcus Rickert

        http://www.the-rickerts.de/mr/PAMINA.html
*/

// ***INCLUDES************************************************************************ 

#include <TBaseStream.h>
#include <TTimeStep.h>
#include <TTransferInfo.h>
#include <TObjectManager.h>
#include <TMultilaneGrid.h>
#include <TRouteplan.h>
#include <TVehicle.h>

#include "TRouter.h"
#include "TSource.h"

// ***GLOBAL VARIABLES****************************************************************  

TClassInfo *TSource::ClassInfo = 0;
int         TSource::InsertionGap = 2;
int         TSource::InsertionSlack = 5;
long        TSource::StartTimeStep = 0;

// ***METHODS************************************************************************* 

TSourceVarData::TSourceVarData(void)

{
    NormalInsertions = 0;
    DelayedInsertions = 0;
    SumDelay = 0;
    PendingVehicles = new TVehicleList;
}

// =================================================================================== 

TSourceVarData::~TSourceVarData(void)

{
    if (PendingVehicles)
    {
        PendingVehicles->DeleteAll();
        DeleteObject(PendingVehicles);
    }    
}

// =================================================================================== 
#define ParentClass TParallelObject
// =================================================================================== 

TSource::TSource(TParallelID theParallelID) : 

ParentClass(ClassInfo, theParallelID)

{
    Grid = 0;
    Offset = INVALID;
}

// =================================================================================== 

TSource::~TSource(void)

{
}

/* =================================================================================== */

bool
TSource::AddVehicle(TVehicle *Vehicle)

{
    Print(DM_ROUTE, "vehicle %d added to list", (char *) Vehicle->GetInstanceName());
    Var.PendingVehicles->Append(Vehicle);
    return TRUE;
}

/* =================================================================================== */

bool
TSource::AssignGrid(TMultilaneGrid *theGrid, long theOffset)

{
    Grid = theGrid;
    Offset = theOffset;
    return TRUE;
}

/* =================================================================================== */

bool
TSource::Decode(TBaseStream *aStream, TTransferInfo *TransferInfo)

{
    if (!ParentClass::Decode(aStream, TransferInfo))
        return FALSE;
    switch(TransferInfo->TransferMode)
    {
    case TM_ConstData:
        return TRUE;

    case TM_VarData:
        return (aStream->Decode(Var.NormalInsertions, TSOURCE_LONG) AND
                DecodeVehicles(aStream));

    default:
        Warning("Decode: invalid TransferMode %d", TransferInfo->TransferMode);
    }
    return TRUE;
}

/* =================================================================================== */

bool
TSource::DecodeVehicles(TBaseStream *aStream)

{
    if (!ObjectManager->ReceiveObjectList((TParallelObjectList *) Var.PendingVehicles,
                                          aStream, FALSE))
    {
        Error("DecodeVehicles: cannot receive");
        return FALSE;
    }
    return TRUE;
}

// =================================================================================== 

bool
TSource::Encode(TBaseStream *aStream, TTransferInfo *TransferInfo) const

{
    if (!ParentClass::Encode(aStream, TransferInfo))
        return FALSE;
    switch(TransferInfo->TransferMode)
    {
    case TM_ConstData:
        return TRUE;
        
    case TM_VarData:
        return (aStream->Encode(Var.NormalInsertions, TSOURCE_LONG) AND
                EncodeVehicles(aStream));

    default:
        Warning("Encode: invalid TransferMode %d", TransferInfo->TransferMode);
    }
    return TRUE;
}

// =================================================================================== 

bool
TSource::EncodeVehicles(TBaseStream *aStream) const

{
    if (!ObjectManager->SendObjectList((TParallelObjectList *) Var.PendingVehicles,
                                       aStream))
    {
        Error("EncodeVehicles: cannot send");
        return FALSE;
    }
    return TRUE;
}

// =================================================================================== 

bool
TSource::ExecuteSubTimeStep(const TTimeStep &TimeStep)

{
    TVehicleEnum  CVehicle;
    long          CLane;
    bool          Inserted;
    long                 i;

    if (Grid == 0)
    {
        Error("ExecuteSubTimeStep: no grid available");
        return FALSE;
    }
    Var.SumDelay += Var.PendingVehicles->GetNrOfElements();
    CVehicle.WalksForwardThrough(Var.PendingVehicles);
    while (CVehicle())
    {
        TRouteplan *Routeplan;

        Routeplan = CVehicle->GetRouteplan();   
        Inserted = FALSE;
        CLane = Grid->GetNrOfLanes() - 1; 
        while (CLane >= 0 AND !Inserted)
        {
            i = 0;
            while (i <= InsertionGap AND Grid->VAccess(CLane, Offset + i) == 0)
                i++;
            if (i == InsertionGap + 1)
            {
                long CurrentTimeStep;
                long Delay;

                CurrentTimeStep = StartTimeStep + TimeStep;
                CVehicle->SetCurrentVelocityInCAUnits(InsertionGap - 1);
                if (!Grid->SetVehicle(CLane, Offset, CVehicle, CVehicle->GetSitePattern()))
                {
                    Error("ExecuteSubTimeStep: cannot set vehicle");
                    return FALSE;
                }
                Print(DM_ROUTE, "vehicle %s inserted", (char *) CVehicle->GetInstanceName());
                if (!Router->HandleNewRouteplan(TimeStep, Routeplan))
                    return FALSE;
                Var.PendingVehicles->Delete(CVehicle);
                Inserted = TRUE;
                Delay = CurrentTimeStep - Routeplan->GetDepartureTimeStep();
                Var.SumDelay -= Delay;
                if (Delay > InsertionSlack)
                    Var.DelayedInsertions++;
                else
                    Var.NormalInsertions++;             
            }
            CLane--;
        }               
        CVehicle++;
    }
    return TRUE;
}

// =================================================================================== 

TString
TSource::GetClassName(void) const

{
    return "TSource";
}

// =================================================================================== 

double
TSource::GetAverageDelay(void) const

{
    long Count;

    if (Var.PendingVehicles AND
        (Count = Var.PendingVehicles->GetNrOfElements()))
        return (double) Var.SumDelay / Count;
    else
        return 0.0;
}

// =================================================================================== 

TString
TSource::GetInstanceName(void) const

{
    char Line[256];
    static char Format[128] = { 0 };

    if (0 == Format[0]) {
        sprintf(Format, "ID:%s pending:%%ld", PARALLEL_ID_PRINTF_FORMAT);
    }
    sprintf(Line, Format, GetParallelID(),
            Var.PendingVehicles? Var.PendingVehicles->GetNrOfElements():0);
    return Line;
}

// =================================================================================== 
#undef ParentClass

