/*      file     : TrafficData/TTrafficSimulator.C
        project  : PAMINA III
        author   : Marcus Rickert
        RCS      : $Id: TTrafficSimulator.C,v 1.13 2005/03/25 12:28:52 mr Exp $

        Copyright (C) 2005 Marcus Rickert

        http://www.the-rickerts.de/mr/PAMINA.html
*/

// ***INCLUDES************************************************************************ 

#include <TTimer.h>
#include <TStream.h>
#include <TConfiguration.h>
#include <TEnvironment.h>

#include <TTimeStep.h>
#include <TNode.h>
#include <TEdge.h>
#include <TIncidenceInfo.h>
#include <TParallel.h>
#include <TObjectManager.h>
#include <TSamplingDescription.h>
#include <TAggregationInfo.h>
#include <TAggregationResult.h>
#include <TStatManager.h>

#include "CAConfig.h"
#include "TrafficDataTypes.h"
#include "TrafficDataConfig.h"
#include "TSimpleIntersection.h"
#include "TCAManager.h"
#include "TTurnCountStat.h"
#include "TTrafficTimers.h"
#include "TTrafficSimulator.h"

// ***DEFINES************************************************************************* 

#define TRAFFIC_TURNCOUNTEXTENSION  "tcs"

#define SIMULATOR_DEFAULT_USETRIVINT   0L

#define TRAFFIC_DEFAULT_TURNCOUNTSTATINTERVAL   -1L

#define INTERSECTION_DEFAULT_SCANRANGE ((long) CA_DEFAULT_V_MAX)

// ***GLOBAL VARIABLES**************************************************************** 

TTrafficSimulator *TrafficSimulator = 0;

// ***METHODS************************************************************************* 

// =================================================================================== 
#define ParentClass TBaseObject
// =================================================================================== 

TTrafficSimulator::TTrafficSimulator(TNodeList *theNodes, TEdgeList *theEdges,
                                     TEdgeList *theBoundaryEdges,
                                     TEdgeList *theNonBoundaryEdges,
                                     TTimer *theNodeTimer, TTimer *theEdgeTimer)

{
    TurnCountStatOutput = 0;
    Nodes = theNodes;
    Edges = theEdges;
    BoundaryEdges = theBoundaryEdges;
    NonBoundaryEdges = theNonBoundaryEdges;
    NodeTimer = theNodeTimer;
    EdgeTimer = theEdgeTimer;
    CAManager = new TCAManager(theEdges, theBoundaryEdges, theNonBoundaryEdges);
    TrafficSimulator = this;
}

// =================================================================================== 

TTrafficSimulator::~TTrafficSimulator(void)

{
    TrafficSimulator = 0;
    DeleteObject(TurnCountStatOutput);
    DeleteObject(CAManager);
}

// =================================================================================== 

bool
TTrafficSimulator::Activate(void)

{
    ObjectManager->RegisterClass(CI_FLAG_TABLE, TSimpleIntersection::ClassInfo, 
                                 CLASS_NODE, CLASS_SIMPLE_INTERSECTION, this);
    ObjectManager->RegisterClass(CI_FLAG_TABLE, TTurnCountStat::ClassInfo, 
                                 0, CLASS_TURN_COUNT_STAT, this);
    Configuration->SetDefault(TConfigEntry(TRAFFIC_GROUP, TRAFFIC_NAME_TURNCOUNTSTATINTERVAL,
                                           TRAFFIC_DEFAULT_TURNCOUNTSTATINTERVAL));
    Configuration->SetDefault(TConfigEntry(INTERSECTION_GROUP, INTERSECTION_NAME_SCANRANGE, 
                                           INTERSECTION_DEFAULT_SCANRANGE));
    return CAManager->Activate();
}

// ===================================================================================

bool
TTrafficSimulator::ActivateStatistics(void)

{
    if (!StatManager->RegisterSamplingHandler(TTurnCountStat::ClassInfo->GetID(), this))
    {
        Error("ActivateStatistics: cannot register myself as sampling handler");
        return FALSE;
    }
    TurnCountStatInterval = (int) Configuration->GetLong(TRAFFIC_GROUP, 
                                                         TRAFFIC_NAME_TURNCOUNTSTATINTERVAL);
    if (TurnCountStatInterval > 0)
    {
        if (Parallel->IsMaster())
        {
            TurnCountSamplingDescription = 
                 new TSamplingDescription(TTurnCountStat::ClassInfo->GetID(), SF_MERGE,
                                          Parallel->GetLocalCPNID(), SAMPLING_ID_TURN_COUNT_STAT);
            TurnCountSamplingDescription->SetCollectInterval(TurnCountStatInterval);
            TurnCountAggregationInfo = new TAggregationInfo(this, TurnCountSamplingDescription);
            StatManager->RegisterAggregation(TurnCountAggregationInfo);
            StatManager->StartSampling(TurnCountSamplingDescription);
        }
    }
    return TRUE;
}

// =================================================================================== 

bool
TTrafficSimulator::Collect(TSamplingDescription *Description)

{
    if (Description->GetStatObjectClassID() == TTurnCountStat::ClassInfo->GetID())
        return CollectTurnCounts(Description->GetTimeStep(), Description->GetObjectList());
    return TStatHandler::Collect(Description);
}

// =================================================================================== 

bool
TTrafficSimulator::CollectTurnCounts(const TTimeStep &TimeStep, TStatObjectList *StatList)

{
    TEnumerator< TIntersectionControl > CControl(TIntersectionControl::ClassInfo->GetTable());
    TInfoID InfoID;

    InfoID = TSimpleIntersection::ClassInfo->GetInfoID();
    CControl.Enumerate();
    while (CControl())
    {
        TTurnCountStat      *Stat;
        TSimpleIntersection *Intersection;

        Intersection = (TSimpleIntersection *) CControl->GetInfo(InfoID);
        Stat = new TTurnCountStat(Intersection->GetParallelID());
        Stat->Collect(Intersection);
        Stat->SetTimeStep(TimeStep);
        StatList->Append(Stat);
        Intersection->ResetTurnCounts();
        CControl++;
    }
    return TRUE;
}

// =================================================================================== 

bool
TTrafficSimulator::ExecuteIntersections(const TTimeStep &TimeStep)

{
    if (TimeStep.GetSubStep() == 0)
    {
        TEnumerator< TSimpleIntersection > CIntersection(TSimpleIntersection::ClassInfo->GetTable());
        
        CIntersection.Enumerate();
        while (CIntersection())
        {
            if (TimeNodesNow(TimeStep))
            {
                TNode *Node;
                
                Node = (TNode *) CIntersection->GetContainer();
                if (!Node->StartTimer(NodeTimerClassInfo->GetInfoID()))
                {
                    Error("ExecuteSubTimeStep: cannot start timer for %s",
                          (char *) Node->GetObjectName());
                    return FALSE;
                }
            }
            if (!CIntersection->ExecuteSubTimeStep(TimeStep))
                return FALSE;
            if (TimeNodesNow(TimeStep))
            {
                TNode *Node;
                
                Node = (TNode *) CIntersection->GetContainer();
                Node->StopTimer(NodeTimerClassInfo->GetInfoID(), 
                                DEFAULT_TIMER_SAMPLE_MASK);
            }
            CIntersection++;
        }
    }
    return TRUE;
}

// =================================================================================== 

bool
TTrafficSimulator::ExecuteIntersectionControls(const TTimeStep &TimeStep)

{
    if (TimeStep.GetSubStep() == 0)
    {
        TEnumerator< TIntersectionControl > 
            CIntersectionControl(TIntersectionControl::ClassInfo->GetTable());
        
        CIntersectionControl.Enumerate();
        while (CIntersectionControl())
        {
            if (TimeNodesNow(TimeStep))
            {
                TNode *Node;
                
                Node = (TNode *) CIntersectionControl->GetContainer();
                if (!Node->StartTimer(NodeTimerClassInfo->GetInfoID()))
                {
                    Error("ExecuteSubTimeStep: cannot start timer for %s",
                          (char *) Node->GetObjectName());
                    return FALSE;
                }
            }
            if (!CIntersectionControl->ExecuteSubTimeStep(TimeStep))
                return FALSE;
            if (TimeNodesNow(TimeStep))
            {
                TNode *Node;
                
                Node = (TNode *) CIntersectionControl->GetContainer();
                Node->StopTimer(NodeTimerClassInfo->GetInfoID(), 
                                DEFAULT_TIMER_SAMPLE_MASK);
            }
            CIntersectionControl++;
        }
    }
    return TRUE;
}

// =================================================================================== 

bool
TTrafficSimulator::ExecuteSubTimeStep(TTimeStep &TimeStep, bool BoundariesAvailable)

{
    if (!BoundariesAvailable)
    {   
        NodeTimer->Select();
        if (!PrepareIntersections(TimeStep))
            return FALSE;
        EdgeTimer->Select();
        if (!CAManager->ExecuteAccessories(TimeStep))
            return FALSE;
        NodeTimer->Select();
        if (!ExecuteIntersections(TimeStep))
            return FALSE;
        if (!ExecuteIntersectionControls(TimeStep))
            return FALSE;
    }
    EdgeTimer->Select();
    if (!CAManager->ExecuteSubTimeStep(TimeStep, BoundariesAvailable))
    {
        Error("ExecuteSubTimeStep: while calling CAManager");
        return FALSE;
    }
    EdgeTimer->Unselect();
    return TRUE;
}

// =================================================================================== 

TString
TTrafficSimulator::GetClassName(void) const

{
    return "TTrafficSimulator";
}

// =================================================================================== 

long
TTrafficSimulator::GetStartTimeStep(void) const

{
    return StartTimeStep;
}

// =================================================================================== 

bool
TTrafficSimulator::HandleAggregationResult(TAggregationInfo *AggregationInfo, 
                                             TTimeStep &TimeStep)

{
    if (AggregationInfo == TurnCountAggregationInfo)
        return HandleTurnCountResult(AggregationInfo, TimeStep);
    return TStatHandler::HandleAggregationResult(AggregationInfo, TimeStep);
}

// =================================================================================== 

bool
TTrafficSimulator::HandleCreateRequest(TParallelObject *&Object, 
                                       const TClassInfo *ClassInfo, 
                                       TParallelID ParallelID)

{
    if (ClassInfo == TSimpleIntersection::ClassInfo)

        Object = new TSimpleIntersection(ParallelID);

    if (ClassInfo == TTurnCountStat::ClassInfo)

        Object = new TTurnCountStat(ParallelID);

    return TRUE;
}

// =================================================================================== 

bool
TTrafficSimulator::HandleTurnCountResult(TAggregationInfo *AggregationInfo, 
                                         TTimeStep &TimeStep)

{
    TAggregationResult   *Result = 0;
    TTurnCountStatList   *StatList = 0;
    TTurnCountStatEnum    CInfo;
                    
    if (TurnCountStatOutput == 0)
    {
        char Filename[512];
        sprintf(Filename, "%s.%s", (char *) Environment->ReplacePathTokens("%R"), 
                TRAFFIC_TURNCOUNTEXTENSION);
        TurnCountStatOutput = new TStream;                  
        if (!TurnCountStatOutput->Open(Environment->ReplacePathTokens(Filename), "w"))
        {
            Error("ActivateStatistics: cannot open edge time stat file %s", Filename);
            return FALSE;
        }
    }
    if ((Result = (TAggregationResult *) AggregationInfo->GetResult(TimeStep)) == 0)
    {
        Error("HandleTurnCountResult: cannot retrieve aggregation result");
        return FALSE;
    }
    if ((StatList = (TTurnCountStatList *) Result->StatObjectList) == 0)
    {
        Error("HandleTurnCountResult: cannot retrieve stat list");
        return FALSE;
    }
    CInfo.WalksForwardThrough(StatList);
    while (CInfo())
    {
        if (!CInfo->Write(TurnCountStatOutput))
        {
            Error("HandeTurnCountResult: cannot write to file");
            return FALSE;
        }
        CInfo++;
    }
    AggregationInfo->DisposeResult(TimeStep);
    return TRUE;
}

// =================================================================================== 

bool
TTrafficSimulator::HandleLoadRequest(TParallelObject *Object, bool &Loaded)

{
    if (Object->GetClassID() == TSimpleIntersection::ClassInfo->GetID())

        Loaded = TRUE;

    return TRUE;
}

// =================================================================================== 

bool
TTrafficSimulator::InitializeEdgeAfterReception(TEdge *Edge, 
                                                TDistributedRange *ActiveRange)

{
    return CAManager->InitializeEdgeAfterReception(Edge, ActiveRange);
}

// =================================================================================== 

bool
TTrafficSimulator::InitializeNodeAfterReception(TNode *Node)

{
    TIncidenceInfo       *IncidenceInfo;
    TIntersection        *Intersection;
    TIntersectionControl *IntersectionControl;
    TInfoID               InfoID;

    if ((IncidenceInfo = (TIncidenceInfo *) 
         Node->GetInfo(TIncidenceInfo::ClassInfo->GetInfoID())) == 0)
    {
        Error("InitializeNodeAfterReception: cannot retrieve incidence info for %s",
              (char *) Node->GetInstanceName());
        return FALSE;
    }
    InfoID = TSimpleIntersection::ClassInfo->GetInfoID();
    if ((Intersection = (TIntersection *) Node->GetInfo(InfoID, TRUE, TRUE)) == 0)
    {
        Error("InitializeNodeAfterReception: cannot create intersection for %s",
              (char *) Node->GetInstanceName());
        return FALSE;
    }
    if (!Intersection->AssignEdges(IncidenceInfo))
    {
        Error("InitializeNodeAfterReception: cannot assign edges to %s",
              (char *) Intersection->GetObjectName());
        return FALSE;
    }
    InfoID = TIntersectionControl::ClassInfo->GetInfoID();
    if ((IntersectionControl = (TIntersectionControl *) Node->GetInfo(InfoID, TRUE, TRUE)))
    {
        if (!IntersectionControl->Activate(IncidenceInfo))
        {
            Error("InitializeNodeAfterReception: cannot activate %s",
                  (char *) IntersectionControl->GetObjectName());
            return FALSE;           
        }
    }
    if (!Intersection->Activate())
    {
        Error("InitializeNodeAfterReception: cannot activate %s",
              (char *) Intersection->GetObjectName());
        return FALSE;
    }
    return TRUE;
}

// =================================================================================== 

bool
TTrafficSimulator::PrepareIntersections(const TTimeStep &TimeStep)

{
    TEnumerator< TSimpleIntersection > CIntersection(TSimpleIntersection::ClassInfo->GetTable());

    CIntersection.Enumerate();
    while (CIntersection())
    {
        if (TimeNodesNow(TimeStep))
        {
            TNode *Node;
            
            Node = (TNode *) CIntersection->GetContainer();
            if (!Node->StartTimer(NodeTimerClassInfo->GetInfoID()))
            {
                Error("ExecuteSubTimeStep: cannot start timer for %s",
                      (char *) Node->GetObjectName());
                return FALSE;
            }
        }
        if (!CIntersection->PrepareSubTimeStep(TimeStep))
            return FALSE;
        if (TimeNodesNow(TimeStep))
        {
            TNode *Node;
            
            Node = (TNode *) CIntersection->GetContainer();
            Node->StopTimer(NodeTimerClassInfo->GetInfoID(), 
                            DEFAULT_TIMER_SAMPLE_MASK);
        }
        CIntersection++;
    }
    return TRUE;
}

// =================================================================================== 

bool
TTrafficSimulator::Run(long theStartTimeStep)

{
    StartTimeStep = theStartTimeStep;
    TSimpleIntersection::ScanRange = 
        (int) Configuration->GetLong(INTERSECTION_GROUP, INTERSECTION_NAME_SCANRANGE);
    if (!ActivateStatistics())
    {
        Error("Run: cannot activate statistics");
        return FALSE;
    }
    return CAManager->Run(StartTimeStep);
}

// =================================================================================== 
#undef ParentClass


