/*      file     : XTrafficData/TMapViewer.C
        project  : PAMINA III
        author   : Marcus Rickert
        RCS      : $Id: TMapViewer.C,v 1.6 2005/03/25 12:29:05 mr Exp $

        Copyright (C) 2005 Marcus Rickert

        http://www.the-rickerts.de/mr/PAMINA.html
*/

// ***INCLUDES************************************************************************ 

#include <ParToolsTypes.h>
#include <TConfiguration.h>
#include <TEnvironment.h>
#include <TObjectManager.h>
#include <TTimeStep.h>
#include <TNetwork.h>
#include <TExtendedCPNInfo.h>
#include <TIncidenceInfo.h>
#include <TNode.h>
#include <TEdge.h>
#include <TGraphEdgeInfo.h>
#include <TGridStatLookup.h>

#include "XTrafficDataConfig.h"

#include "TMapView.h"
#include "TMapViewer.h"

// ***DEFINES************************************************************************* 

#define MAPWIN_DEFAULT_NAME               "view"
#define MAPWIN_DEFAULT_WIDTH              300L
#define MAPWIN_DEFAULT_HEIGHT             300L
#define MAPWIN_DEFAULT_DUMPINTERVAL       -1L

// ***METHODS************************************************************************* 

// =================================================================================== 
#define ParentClass TBaseObject
// =================================================================================== 

TMapViewer::TMapViewer(TRectangle *theFrame, TNodeList *theNodes, TEdgeList *theEdges)

{
    TopologyChanged = TRUE;
    Frame = theFrame;
    Nodes = theNodes;
    Edges = theEdges;
    DumpInterval = Configuration->Exists(MAPWIN_GROUP, MAPWIN_NAME_DUMPINTERVAL)? 
        Configuration->GetLong() : MAPWIN_DEFAULT_DUMPINTERVAL;
    Name = Configuration->Exists(MAPWIN_GROUP, MAPWIN_NAME_NAME)? 
        Configuration->GetString() : MAPWIN_DEFAULT_NAME;
    Width = Configuration->Exists(MAPWIN_GROUP, MAPWIN_NAME_WIDTH)? 
        Configuration->GetLong() : MAPWIN_DEFAULT_WIDTH;
    Height = Configuration->Exists(MAPWIN_GROUP, MAPWIN_NAME_HEIGHT)? 
        Configuration->GetLong() : MAPWIN_DEFAULT_HEIGHT;
    GridStatLookupClassInfo->RegisterActionHandler(this);
}

// =================================================================================== 

TMapViewer::~TMapViewer(void)

{
    CloseViews(&Views);
}

// =================================================================================== 

bool
TMapViewer::CheckOpen(void)

{
    if (Views.GetNrOfElements() == 0)
    {
        if (!ColorCPN())
        {
            Error("CheckOpen: cannot color CPN");
            return FALSE;
        }
        if (!OpenView(*Frame))
        {
            Error("CheckOpen: cannot open main view");
            return FALSE;
        }
    }
    return TRUE;
}

// =================================================================================== 

void
TMapViewer::CloseView(TMapView *View)

{
    ViewsToBeClosed.Append(View);   
}

// =================================================================================== 

void
TMapViewer::CloseViews(TMapViewList *List)

{
    TMapViewEnum CView;
    
    if (ViewsToBeClosed.GetNrOfElements() > 0)
    {
        CView.WalksForwardThrough(List);
        while (CView())
        {
            TMapView *View;
            
            View = CView;
            Print("closing %s", (char *) View->GetObjectName());
            DeleteObject(View);
            CView++;
        }
        MarkViews();
    }
}

// =================================================================================== 

bool
TMapViewer::ColorEdge(TEdge *Edge)

{
    TExtendedCPNInfo *FromInfo, *ToInfo = 0;
    TGraphEdgeInfo   *GraphInfo;
    TInfoID           InfoID;

    InfoID = TGraphEdgeInfo::ClassInfo->GetInfoID();
    if ((GraphInfo = (TGraphEdgeInfo *) Edge->GetInfo(InfoID, TRUE, TRUE)) == 0)
    {
        Error("ColorEdges: cannot retrieve TGraphEdgeInfo");
        return FALSE;       
    }
    if ((FromInfo = Network->GetExtendedCPNInfo(Edge->GetFromNodeCPNID(), TRUE)) AND
        (ToInfo = Network->GetExtendedCPNInfo(Edge->GetToNodeCPNID(), TRUE)))
    {
        GraphInfo->FromColor = FromInfo->Color;
        GraphInfo->ToColor = ToInfo->Color;
    }
    return TRUE;
}

// =================================================================================== 

bool
TMapViewer::ColorEdges(void)

{
    TEdgeEnum CEdge;
    CEdge.WalksForwardThrough(Edges);
    while (CEdge())
    {
        if (!ColorEdge(CEdge))
        {
            Error("ColorEdges: cannot color %s",  (char *) CEdge->GetObjectName());
            return FALSE;
        }
        CEdge++;
    }
    return TRUE;
}

// =================================================================================== 

bool
TMapViewer::ColorCPN(void)

{
    TEnumerator < TExtendedCPNInfo > CCPN(TExtendedCPNInfo::ClassInfo->GetTable());
    TInfoID InfoID;
    int     CColor;

    InfoID = TGraphEdgeInfo::ClassInfo->GetInfoID();
    CCPN.Enumerate();
    CColor = 0;
    while (CCPN())
    {
        CCPN->Color = - (1 + (CColor % 8));
        CColor++;
        CCPN++;
    }
    return TRUE;
}

// =================================================================================== 

bool
TMapViewer::ExecuteSubTimeStep(const TTimeStep &TimeStep)

{
    TMapViewEnum CView;
    
    if (!CheckOpen())
    {
        Error("ExecuteSubTimeStep: cannot open");
        return FALSE;
    }
    CloseViews(&ViewsToBeClosed);
    if (TopologyChanged)
    {
        if (!ColorEdges())
        {
            Error("CheckOpen: cannot color edges");
            return FALSE;
        }
        TopologyChanged = FALSE;
    }
    CView.WalksForwardThrough(Views);
    while (CView())
    {
        if (DumpInterval != INVALID)
            if (((long) TimeStep) % DumpInterval == 0)
            {
                CView->SetDumpBaseFilename(Environment->ReplacePathTokens("%R"));
                CView->DumpOnNextRedraw();
            }
        if (!CView->Execute(TimeStep))
        {
            Error("Draw: cannot execute view %s", 
                  (char *) CView->GetInstanceName());
            return FALSE;
        }
        CView++;
    }
    return TRUE;
}

// =================================================================================== 

TString
TMapViewer::GetClassName(void) const

{
    return "TMapViewer";
}

// =================================================================================== 

TString
TMapViewer::GetInstanceName(void) const

{
    char Line[128];

    sprintf(Line, "Views:%ld", Views.GetNrOfElements());
    return Line;
}

// =================================================================================== 

TMapViewList *
TMapViewer::GetMapViewList(void)

{
    return &Views;
}

// =================================================================================== 

bool
TMapViewer::HandleAction(TParallelObject *, TObjectAction)

{
    return TRUE;
}

// =================================================================================== 

bool
TMapViewer::HandleAnonymousAction(TClassInfo *, TObjectAction Action)

{
    switch(Action)
    {
    case ActionModified:

        MarkViews();
        break;

    default:
        ;
    }
    return TRUE;
}

// =================================================================================== 

void
TMapViewer::MarkViews(void)

{
    TMapViewEnum CView;
    
    CView.WalksForwardThrough(Views);
    while (CView())
    {
        CView->MarkForRedraw();
        CView++;
    }
}

// =================================================================================== 

bool
TMapViewer::OpenView(TRectangle Frame)

{
    TMapView *MapView;
    double    Ratio;
    int       NewHeight, NewWidth;
    
    Ratio = Frame.GetDeltaX() / Frame.GetDeltaY();
    NewWidth = Width;
    NewHeight = Height;
    if (Ratio > NewWidth / NewHeight)
        NewHeight = (int) (1.0 * NewWidth / Ratio);
    else
        NewWidth = (int) (Ratio * NewHeight);

    Print("OpenView: Frame %s Width:%d Height:%d",
          (char *) Frame.GetInstanceName(), NewWidth, NewHeight);

    MapView = new TMapView(this, "View", &Frame, Edges, NewWidth, NewHeight);
    if (MapView->Failed())
    {
        Error("OpenView: cannot map view");
        return FALSE;
    }
    Views.Append(MapView);
    MarkViews();
    return TRUE;
}

// =================================================================================== 

bool
TMapViewer::UpdateNode(TNode *)

{
    TopologyChanged = TRUE;
    return TRUE;
}

// =================================================================================== 
#undef ParentClass
