/*      file     : common/TABTree.C
        project  : General Toolbox
        author   : Marcus Rickert
        RCS      : $Id: TABTree.C,v 1.11 2005/03/25 11:04:59 mr Exp $

        Copyright (C) 2005 Marcus Rickert

        http://www.the-rickerts.de/mr/PAMINA.html
*/

// ***INCLUDES************************************************************************

#include <stdio.h>
#include <stdlib.h>

#include "TDebug.h"
#include "TString.h"
#include "TEnumerator.h"
#include "TSelector.h"
#include "TABTreeElement.h"
#include "TABTree.h"

// ***METHODS*************************************************************************

// =================================================================================== 
#define ParentClass TContainer
// =================================================================================== 

TABTreePrimitive::TABTreePrimitive(long theFlags) :

ParentClass(theFlags)
    
{
    Init();
    debug(Print(DM_AB_TREE, "CONSTRUCTOR"));
}

// ===================================================================================

TABTreePrimitive::~TABTreePrimitive(void)

{
    DumpEntries();
    if (AutoDisposeEntries())
        DeleteAll();
}

// ===================================================================================

bool
TABTreePrimitive::Apply(TApplyFunction *ApplyFunction, void *Context)

{
    if (Root)
        return Root->Apply(ApplyFunction, Context);
    else
        return TRUE;
}

// ===================================================================================

bool
TABTreePrimitive::Check(void)

{
    Print("Checking consistency...");
    if (Root)
        return Root->Check(0);
    else
        return TRUE;
}

// ===================================================================================

TBaseObject *
TABTreePrimitive::_Delete(TABTreeElement *Element, int DeletionPoint)

{
    TABTreeElement *LeftNeighbour;
    TABTreeElement *RightNeighbour;
    TABTreeElement *Father;
    TBaseObject    *DeletedItem = 0;

    DeletedItem = Element->GetSon(DeletionPoint);
    Element->Remove(DeletionPoint);
    
    switch(--NrOfElements)
    {
    case 0:

        // tree is now empty -> delete root
        
        delete Root;
        Root = NULL;
        break;
        
    case 1:

        // root is allowed to have one son only
        
        break;
        
    default:
        while (Element->TooFewSons())
        {
            IfDebug(DM_AB_TREE, Print("Current element:%s root:%s", 
                                      (char *) Element->GetInstanceName(),
                                      (Element == Root)? "yes":"no"));
            if ((LeftNeighbour = Element->GetLeftNeighbour()) AND 
                LeftNeighbour->DeletePossible())
            {
                // Steal from left neighbour

                IfDebug(DM_AB_TREE, Print("Steal from left neighbour"));
                Element->Insert(0, LeftNeighbour->StealRightmostSon());
                break;
            }
            else
            {
                if ((RightNeighbour = Element->GetRightNeighbour()) AND
                    RightNeighbour->DeletePossible())
                {
                    // Steal from right neighbour

                    IfDebug(DM_AB_TREE, Print("Steal from right neighbour"));
                    Element->Insert(Element->GetElementsUsed(), 
                                    RightNeighbour->StealLeftmostSon());
                    break;
                }
                else
                {
                    Father = Element->GetFather();
                    if (LeftNeighbour)
                    {
                        // Merge element with left neighbour,
                        // current element is deleted

                        IfDebug(DM_AB_TREE, Print("Merge with left neighbour. Father: %s",
                                         (char *) Father->GetInstanceName()));
                        LeftNeighbour->Merge(Element);
                        Father->Delete(Father->GetSonNumber(Element));
                        Element = Father;
                    }
                    else
                    {
                        if (RightNeighbour)
                        {
                            // Merge element with right neighbour,
                            // right neighbour is deleted

                            IfDebug(DM_AB_TREE, Print("Merge with right neighbour"));
                            Element->Merge(RightNeighbour);
                            Father->Delete(Father->GetSonNumber(RightNeighbour));
                            Element = Father;
                        }
                        else
                        {
                            // if there is neither a left nor a right
                            // neighbour, we are at the root!
                            // delete the old root and make the current
                            // element the new root

                            IfDebug(DM_AB_TREE, Print("Delete root"));
                            Element = Element->GetSon(0);
                            Root->Reset();
                            delete Root;
                            Root = Element;
                            Root->SetFather(0);
                            break;
                        }
                    }
                }
            }
            Element->RetrieveLeftmostItem();
        }
        Element->TraceLeftmostItem();
    }
#ifdef DTREE
    if (!Check())
        Error("Inconsistency after DELETE");
#endif
    return DeletedItem;
}

// ===================================================================================

TBaseObject *
TABTreePrimitive::DeleteByValue(TBaseObject *Item, int CompareCode)

{
    TABTreeElement *Element;
    int             DeletionPoint;
    bool            Found;

    if (Root == 0)
        return 0;
    Element = FindLeaf(Item, CompareCode, FALSE);
    DeletionPoint = FindItem(Element, Item, Found, CompareCode, FALSE);
    if (!Found)
        return 0;
    return _DisposeContainerEntry(_Delete(Element, DeletionPoint));
}

// ===================================================================================

TBaseObject *
TABTreePrimitive::_DeleteByReference(TBaseObject *Item, int CompareCode)

{
    TABTreeElement *Element;
    int             DeletionPoint;
    bool            Found;

    if (Root == 0)
        return 0;
    Element = FindLeaf(Item, CompareCode, TRUE);
    DeletionPoint = FindItem(Element, Item, Found, CompareCode, TRUE);
    if (!Found)
        return 0;
    return _DisposeContainerEntry(_Delete(Element, DeletionPoint));
}

// ===================================================================================

void
TABTreePrimitive::DeleteAll(void)
    
{
    if (Root != 0)
    {
        if (AutoDisposeObjects())
            Root->DeleteObjects();
        delete Root;
        Init();
    }
}

// ===================================================================================

void
TABTreePrimitive::Enumerate(TEnumeratorPrimitive *Enumerator)

{
    TABTreeElement *DataPtr = 0;

    if (Root != 0)
        DataPtr = FindLeftMostLeaf();
    Enumerator->SetDataPtr(DataPtr);
    Enumerator->SetIndex(0);
}

// ===================================================================================

TBaseObject *
TABTreePrimitive::_Find(TBaseObject *Element, int CompareCode)

{
    TABTreeElement *Current;
    int             CurrentItem;
    bool            Found = FALSE;

    if (Root == 0)
        return 0;
    Current = FindLeaf(Element, CompareCode, FALSE);
    CurrentItem = FindItem(Current, Element, Found, CompareCode, FALSE);
    if (Found)
        return (TBaseObject *) Current->GetCompareItem(CurrentItem);
    else
        return 0;
    }

// ===================================================================================

int
TABTreePrimitive::FindItem(TABTreeElement *Node, TBaseObject *Item, bool& Found,
                           int CompareCode, bool CheckReference)

{
    int CurrentItem;
    int Result = 0;

    /*
    Print("FindItem: '%s'", (char *) Item->GetInstanceName());
    for (CurrentItem = 0; CurrentItem < Node->GetElementsUsed(); CurrentItem++)
        Print("Entry: '%s'", (char *) Node->GetCompareItem(CurrentItem)->GetInstanceName());
    Print("");
    */

    Found = FALSE;
    /*
    CurrentItem = 0;
    while (CurrentItem < Node->GetElementsUsed() AND
           (Result = CompareItems(Item, Node->GetCompareItem(CurrentItem), 
                                  CompareCode, CheckReference)) >= 0)
    {
        if (Result == 0)
        {
            Found = TRUE;
            break;
        }
        CurrentItem++;
    }
    */
    CurrentItem = Node->GetElementsUsed() - 1;
    while (CurrentItem >= 0 AND
           (Result = CompareItems(Item, Node->GetCompareItem(CurrentItem), 
                                  CompareCode, CheckReference)) < 0)
        CurrentItem--;
    if (Result == 0)
        Found = TRUE;
    else
        CurrentItem++;
    IfDebug(DM_AB_TREE, Print("FindItem returns CurrentItem %d", CurrentItem));
    return CurrentItem;
    }

// ===================================================================================

TABTreeElement *
TABTreePrimitive::FindLeaf(TBaseObject *Element, int CompareCode, bool CheckReference)

{
    TABTreeElement *Current;
    int             CurrentSon;
    
    Current = Root;
    while (!Current->IsLeaf())
    {
        CurrentSon = Current->GetElementsUsed() - 1;
        while(CurrentSon >= 0  AND
              CompareItems(Element, Current->GetCompareItem(CurrentSon), 
                           CompareCode, CheckReference) < 0)
            CurrentSon--;
        if (CurrentSon < 0)
            CurrentSon = 0;
        Current = Current->GetSon(CurrentSon);
        IfDebug(DM_AB_TREE, Print("Branch to son %d", CurrentSon));
    }
    return Current;
 }

// ===================================================================================

TABTreeElement *
TABTreePrimitive::FindLeftMostLeaf(void)

{
    TABTreeElement *Current;
    
    Current = Root;
    while (!Current->IsLeaf())
        Current = Current->GetSon(0);
    return Current;
}

// ===================================================================================

TString
TABTreePrimitive::GetClassName(void) const

{
    return "TABTreePrimitive";
}

// ===================================================================================

TBaseObject *
TABTreePrimitive::GetCurrent(TEnumeratorPrimitive *Enumerator)

{
    TABTreeElement *TreeElement = (TABTreeElement *) Enumerator->GetDataPtr();

    if (TreeElement == 0)
        return 0;
    return TreeElement->GetSon((int) Enumerator->GetIndex());
}

// ===================================================================================

TString
TABTreePrimitive::GetInstanceName(void) const

{
    char line[128];

    sprintf(line, "#Elements:%d, Root:%s", NrOfElements, Root? "non-NULL":"NULL");
    return line;
}

// ===================================================================================

long
TABTreePrimitive::GetNrOfElements(void) const

{
    return NrOfElements;
}

// ===================================================================================

TABTreeElement *
TABTreePrimitive::GetRoot(void) const

{
    return Root;
}

// ===================================================================================

void
TABTreePrimitive::Init(void)

{
    Root = 0;
    NrOfElements = 0;
}

// ===================================================================================

bool
TABTreePrimitive::_Insert(TBaseObject *Item, bool Unique, int CompareCode)

{
    TABTreeElement *Element;
    TABTreeElement *NewElement = NULL;
    TABTreeElement *NewRoot;
    TBaseObject    *ChildElement;
    bool            Found;
    int             InsertionPoint;

    if (Root == 0)
    {
        if ((Root = new TABTreeElement(0, TRUE)) == 0)
            return FALSE;
        Root->Insert(0, Item);
    }
    else
    {
        Element = FindLeaf(Item, CompareCode, TRUE);
        ChildElement = Item;
        do
        {
            InsertionPoint = FindItem(Element, Item, Found, CompareCode, TRUE);
            if (Unique AND Found)
                return FALSE;
            if (Element->InsertPossible())
            {
                Element->Insert(InsertionPoint, ChildElement);
                break;
            }
            else
            {
                if ((NewElement = new TABTreeElement(Element->GetFather(),
                                                     Element->IsLeaf())) == 0)
                    return FALSE;
                Element->Split(NewElement);
                if (InsertionPoint < TAB_SPLIT)

                    // insert into left old son
                    
                    Element->Insert(InsertionPoint, ChildElement);

                else

                    // insert into right new son
                    
                    NewElement->Insert(InsertionPoint-TAB_SPLIT, ChildElement);

                ChildElement = NewElement;
                Item = NewElement->GetLeftMostItem();
                Element = Element->GetFather();
            }
        }
        while(Element != NULL);
        if (Element == NULL)
        {
            // insert new root at top with two sons:
               
            //   left son : old root
            //   right son : new element
            
            if ((NewRoot = new TABTreeElement(0, FALSE)) == 0)
                return FALSE;
            NewRoot->Insert(0, Root);
            NewRoot->Insert(1, NewElement);
            NewElement->SetFather(NewRoot);
            Root->SetFather(NewRoot);
            Root = NewRoot;
        }
        else
            
            // promote a pointer to the leftmost leaf to the root of the subtree

            Element->TraceLeftmostItem();
    }
    NrOfElements++;
#ifdef DTREE
    if (!Check())
        Error("Inconsistency after INSERT");
#endif
    return TRUE;
}

// ===================================================================================

bool
TABTreePrimitive::WhileSelection(TSelectorPrimitive *Selector)

{
    TABTreeElement *TreeElement = (TABTreeElement *) Selector->GetDataPtr();
    
    if (Selector->GetDataPtr())
    {
        /*
        Print("WhileSelection comparing (%s:%s) and (%s:%s)",
              (char *) Selector->GetKeyPtr()->GetClassName(),
              (char *) Selector->GetKeyPtr()->GetInstanceName(),
              (char *) TreeElement->GetCompareItem(Selector->GetIndex())->GetClassName(),
              (char *) TreeElement->GetCompareItem(Selector->GetIndex())->GetInstanceName());
              */
        return CompareItems(Selector->GetKeyPtr(), 
                            TreeElement->GetCompareItem((int) Selector->GetIndex()), 
                            Selector->GetCompareCode(), FALSE) == 0; // was: >=
    }
    else
        return FALSE;
}

// ===================================================================================

bool
TABTreePrimitive::WhileEnumeration(TEnumeratorPrimitive *Enumerator)

{
    return Enumerator->GetDataPtr() != 0;
}

// ===================================================================================

void
TABTreePrimitive::MoveBackward(TEnumeratorPrimitive *Enumerator)

{
    TABTreeElement *TreeElement = (TABTreeElement *) Enumerator->GetDataPtr();

    if (TreeElement)
    {
        if (Enumerator->GetIndex() == 0)
        {
            TABTreeElement *LeftNeighbor;

            Enumerator->SetDataPtr(LeftNeighbor = TreeElement->GetPrev());
            if (LeftNeighbor)
                Enumerator->SetIndex(LeftNeighbor->GetElementsUsed() - 1);
        }
        else
            Enumerator->SetIndex(Enumerator->GetIndex() - 1);
    }
}

// ===================================================================================

void
TABTreePrimitive::MoveForward(TEnumeratorPrimitive *Enumerator)

{
    TABTreeElement *TreeElement = (TABTreeElement *) Enumerator->GetDataPtr();

    if (TreeElement)
    {
        if (Enumerator->GetIndex() == TreeElement->GetElementsUsed() - 1)
        {           
            Enumerator->SetDataPtr(TreeElement->GetNext());
            Enumerator->SetIndex(0);
        }
        else
        {
            Enumerator->SetIndex(Enumerator->GetIndex() + 1);
        }
    }
}

// ===================================================================================

void
TABTreePrimitive::Select(TSelectorPrimitive *Selector)

{
    long            Index = 0;
    TABTreeElement *DataPtr = 0;

    if (Root != 0)
    {
        if ((DataPtr = FindLeaf(Selector->GetKeyPtr(), Selector->GetCompareCode(), FALSE)))
        {
            bool Found = FALSE;
            
            Index = FindItem(DataPtr, Selector->GetKeyPtr(), Found,
                             Selector->GetCompareCode(), FALSE);
            if (Found)
            {
                
            }
            else
                DataPtr = 0;
        }
    }
    Selector->SetDataPtr(DataPtr);
    Selector->SetIndex(Index);
    Selector->TraceLeft();
}

// ===================================================================================
#undef ParentClass

