/*      file     : common/TAsciiDataTypeReader.C
        project  : General Toolbox
        author   : Marcus Rickert
        RCS      : $Id: TAsciiDataTypeReader.C,v 1.18 2005/03/29 20:32:32 mr Exp $

        Copyright (C) 2005 Marcus Rickert

        http://www.the-rickerts.de/mr/PAMINA.html
*/

// ***INCLUDES************************************************************************ 

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>

#include <File.h>
#include <TDebug.h>
#include <TEnvironment.h>

#include <TAsciiDataTypeReader.h>

// ***DEFINES************************************************************************* 

#define INDEX_CREATE_MASK (S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH)

#define DEFAULT_INDEX_SCAN_PATTERN "%lld"

// ***FUNCTIONS*********************************************************************** 

int
CompareIndexEntries(TIndexEntry *Entry1, TIndexEntry *Entry2)

{
    if (Entry1->Key < Entry2->Key)
        return -1;
    else
        if (Entry1->Key > Entry2->Key)
            return 1;
        else
            return 0;
}

// ***METHODS************************************************************************* 

// =================================================================================== 
#define ParentClass TDataTypeReader
// =================================================================================== 

TAsciiDataTypeReader::TAsciiDataTypeReader(const char *theBaseFilename,
                                           const char *theTableName,
                                           const char *theIndexScanPattern,
                                           const char *theIndexFileExtension)

{
    BaseFilename = theBaseFilename;
    TableName = theTableName;
    IndexFileExtension = theIndexFileExtension;
    if (theIndexScanPattern)
        IndexScanPattern = theIndexScanPattern;
    else
        IndexScanPattern = DEFAULT_INDEX_SCAN_PATTERN;
    Input = 0;
    IndexFile = -1;
    ReadOnly = TRUE;
    Indices = 0;
    LinearIndex = 0;
    CurrentSelectionCount = 0;
    PreviousID = INVALID;
    PreviousObject = 0;
}

// =================================================================================== 

TAsciiDataTypeReader::~TAsciiDataTypeReader(void)

{
    DeleteObjectArray(LinearIndex);
    Close();
}

// =================================================================================== 

void
TAsciiDataTypeReader::Close(void)

{
    UnmapIndices();
    if (Input != 0)
    {
        fclose(Input);
        
        Print(DM_FILE, "data file closed.");
        Input = 0;
    }
    if (IndexFile != -1)
    {
        close(IndexFile);
        Print(DM_FILE, "index file closed.");
        IndexFile = -1;
    }
    if (PreviousObject)
    {
        DisposeObject(PreviousObject);
        PreviousObject = 0;
        PreviousID = INVALID;
    }
}

// =================================================================================== 

bool
TAsciiDataTypeReader::CountEntries(void)

{
    if (!Open())
    {
        Error("CountEntries: cannot open input");
        return FALSE;
    }
    NrOfEntries = 0;
    while (!Eof())
        NrOfEntries++;
    Print(DM_FILE, "Counted %i entries for file %s", NrOfEntries, (char *) BaseFilename);
    return TRUE;
}

// =================================================================================== 

bool
TAsciiDataTypeReader::CreateIndices(const char *Filename)

{
    if ((IndexFile = open(Filename, O_RDWR | O_CREAT, INDEX_CREATE_MASK)) == -1)
        Warning("cannot create index file '%s'-> will not save index", Filename);
    else
    {
        Print(DM_FILE, "created index file %s.", Filename);
        ReadOnly = FALSE;
    }
    return CountEntries();
}

// =================================================================================== 

void *
TAsciiDataTypeReader::CreateObject(const char *) const

{
    NotOverloaded("CreateObject");
    return FALSE;
}

// =================================================================================== 

void
TAsciiDataTypeReader::DisposeObject(void *)

{
    NotOverloaded("DisposeObject");
}

// =================================================================================== 

bool
TAsciiDataTypeReader::Eof(void) 

{
    return !ReadLine();
}

// =================================================================================== 

bool
TAsciiDataTypeReader::GenerateIndices(void) 

{
    long i;

    rewind(Input);
    i = 0;
    while (!Eof())
    {
        sscanf(LineBuffer, (char *) IndexScanPattern(), &Indices[i].Key);

        Indices[i].Offset = LineOffset;

        //	Print(DM_FILE, "Index=%s, tell=%ld, Key=%lld, Offset=%lld", (const char *) BaseFilename(), 
        //	      LineOffset, Indices[i].Key, Indices[i].Offset);
        i++;
    }
    Print(DM_CONTROL, "sorting indices.");
    qsort((void *) Indices, NrOfEntries, sizeof(TIndexEntry),
          (int (*) (const void *, const void *)) CompareIndexEntries);
    return TRUE;
}

// =================================================================================== 

void *
TAsciiDataTypeReader::Get(void) const

{
    return CreateObject(LineBuffer);
}

// =================================================================================== 

void *
TAsciiDataTypeReader::Get(LongID)

{
    Warning("Get(long) not implemented yet!");
    return 0;
}

// =================================================================================== 

void *
TAsciiDataTypeReader::GetBuffered(LongID ID)

{
    long Count;

    if (PreviousObject)
    {
        if (PreviousID == ID)
            return PreviousObject;
        DisposeObject(PreviousObject);
        PreviousObject = 0;
    }
    if (!Select(ID, Count))
    {
        Error("GetBuffered: cannot select");
        return 0;
    }
    if (Count != 1)
    {
        Warning("GetBuffered: select returns count=%ld", Count);
        return 0;
    }
    PreviousID = ID;
    return (PreviousObject = GetSelected(0));
}

// =================================================================================== 

TString
TAsciiDataTypeReader::GetClassName(void) const

{
    return "TAsciiDataTypeReader";
}

// =================================================================================== 

TString
TAsciiDataTypeReader::GetInstanceName(void) const

{
    char Line[256];

    sprintf(Line, "%s", TableName());
    return Line;
}

// =================================================================================== 

bool
TAsciiDataTypeReader::GetLinearIndex(LongID *&Index, long &Count)

{
    int i;

    if (LinearIndex == 0)
    {
        if (!PrepareRandomAccess())
        {
            Error("GetLinearIndex: cannot retrieve index");
            return FALSE;
        }
        if ((LinearIndex = new LongID[NrOfEntries]) == 0)
        {
            Error("GetLinearIndex: out of memory");
            return FALSE;
        }
        for (i = 0; i < NrOfEntries; i++)
            LinearIndex[i] = Indices[i].Key;
    }
    Index = LinearIndex;
    Count = NrOfEntries;
    return TRUE;
}

// =================================================================================== 

long
TAsciiDataTypeReader::GetNrOfEntries(void) 

{
    QueryIndices(TRUE);
    return NrOfEntries;
}

// =================================================================================== 

bool
TAsciiDataTypeReader::GetRange(LongID Key, long &First, long &Last)

{
    TIndexEntry  SearchEntry;
    TIndexEntry *Element = 0;

    if (Indices == 0)
    {
        Error("no index available");
        return FALSE;
    }
    SearchEntry.Key = Key;
    Element = (TIndexEntry *) bsearch((void *) &SearchEntry, (void *) Indices, 
              NrOfEntries, sizeof(TIndexEntry),
              (int (*) (const void *, const void *)) CompareIndexEntries);
    if (Element == 0)
        return FALSE;
    First = Last = Element - Indices;
    while (First >= 0 AND Indices[First].Key == Key)
        First --;
    First++;
    while (Last < NrOfEntries AND Indices[Last].Key == Key)
        Last++;
    Last--;
    return TRUE;
}

// =================================================================================== 

void *
TAsciiDataTypeReader::GetSelected(long SelectedEntry)

{
    if (SelectedEntry < 0 OR SelectedEntry > CurrentSelectionCount)
    {
        Error("GetSelected: index %ld of range [0:%ld[", CurrentSelectionCount);
        return 0;
    }
    if (fseek(Input, Indices[FirstEntry + SelectedEntry].Offset, SEEK_SET) == -1)
    {
        Error("cannot lseek to offset %ld", Indices[FirstEntry + SelectedEntry].Offset);
        return 0;
    }
    if (!ReadLine())
    {
        Error("cannot read entry #%ld in selection", SelectedEntry + 1);
        return 0;
    }
    return CreateObject(LineBuffer);
}

// =================================================================================== 

bool
TAsciiDataTypeReader::MapIndices(void)

{
#ifdef NO_MMAP
    if ((Indices = (TIndexEntry *) malloc(NrOfEntries *sizeof(TIndexEntry))) == 0)
    {
        Error("cannot allocate memory for index");
        close(IndexFile);
        IndexFile = -1;
        return FALSE;
    }
    if (IndexFile != -1 AND ReadOnly)
    {
        Print(DM_FILE, "reading indices.");
        read(IndexFile, (char *) Indices, NrOfEntries * sizeof(TIndexEntry));
    }
#else
    Warning("MapIndices not implemented");
#endif
    return TRUE;
}

// =================================================================================== 

bool
TAsciiDataTypeReader::Open(void) 

{
    if (Input == 0)
    {
        if ((Input = fopen(BaseFilename, "r")) == 0)
        {
            Error("cannot open file '%s'", BaseFilename());
            return FALSE;
        }
        Print(DM_FILE, "data file opened.", BaseFilename());
    }
    rewind(Input);
    return TRUE;
}

// =================================================================================== 

bool
TAsciiDataTypeReader::OpenIndices(const char *Filename)

{
    long NrOfBytes;

    if ((IndexFile = open(Filename, O_RDONLY)) == 0)
    {
        Error("cannot open indexfile '%s'", Filename);
        return FALSE;
    }
    ReadOnly = TRUE;
    NrOfBytes = fsize(IndexFile);
    if (NrOfBytes % sizeof(TIndexEntry) != 0)
    {
        Error("file '%s' is not an index file");
        close(IndexFile);
        IndexFile = -1;
        return FALSE;
    }
    NrOfEntries = NrOfBytes / (long) sizeof(TIndexEntry);
    Print(DM_FILE, "opened index file (%ld entries, READ-ONLY).", NrOfEntries);
    return TRUE;
}

// =================================================================================== 

bool
TAsciiDataTypeReader::PrepareLinearAccess(void)

{
    if (!Open())
        return FALSE;
    LineNr = 0;
    return TRUE;
}

// =================================================================================== 

bool
TAsciiDataTypeReader::PrepareRandomAccess(void)

{
    LineNr = -1;
    return (QueryIndices(TRUE) AND
            Open());
}

// =================================================================================== 

bool
TAsciiDataTypeReader::QueryIndices(bool Create)

{
    char    IndexFilename[512];
    TString ArchIndexFilename;
    long    DataFileDate;
    long    IndexFileDate;
    bool    UseExisting = FALSE;

    if (Indices)
        return TRUE;

    sprintf(IndexFilename, "%s.%%A%s", BaseFilename(), IndexFileExtension());
    ArchIndexFilename = Environment->GetFilename(0, IndexFilename);
    if (fexist(ArchIndexFilename) AND fsize(ArchIndexFilename) > 0)
    {
        DataFileDate = fmodified(Environment->GetFilename(0, BaseFilename()));
        IndexFileDate = fmodified(ArchIndexFilename());
        if (IndexFileDate < DataFileDate)
            Print("index file out of date -> rebuild");
        else
            UseExisting = TRUE;
    }
    if (UseExisting)
        return (OpenIndices(ArchIndexFilename) AND
                MapIndices());
    if (!Create)
        return FALSE;
    return (CreateIndices(ArchIndexFilename) AND
            MapIndices() AND
            GenerateIndices());
}

// =================================================================================== 

bool
TAsciiDataTypeReader::ReadLine(void)

{
    bool  Found = FALSE;
    char *S;

    while (!(feof(Input) OR Found))
    {
        LineNr++;
        LineBuffer[0] = 0;
        LineOffset = ftell(Input);
        S = fgets(LineBuffer, MAX_LINE_LENGTH, Input);
        if (S != 0 AND LineBuffer[0] != 0)
            Found = TRUE;
    }
    return Found;
}

// =================================================================================== 

bool
TAsciiDataTypeReader::Select(LongID Key, long &Count)

{
    long LastEntry;

    Count = 0;
    if (!GetRange(Key, FirstEntry, LastEntry))
        return TRUE;
    CurrentSelectionCount = Count = LastEntry - FirstEntry + 1;
    return TRUE;
}

// =================================================================================== 

bool
TAsciiDataTypeReader::UnmapIndices(void)

{
#ifdef NO_MMAP
    long Count;

    if (Indices != 0)
    {
        if (IndexFile != -1 AND !ReadOnly)
        {
            Print(DM_FILE, "writing indices.");
            Count = (long) sizeof(TIndexEntry) * NrOfEntries;
            if (write(IndexFile, (char *) Indices, Count) == -1)
            {
                Error("while writing indices");
                return FALSE;
            }
        }
        free(Indices);
        Indices = 0;
    }
#else
    
#endif
    return TRUE;
}

// =================================================================================== 
#undef ParentClass
