/*      file     : common/TStatistics
        project  : General Toolbox
        author   : Marcus Rickert
        RCS      : $Id: TStatistics.C,v 1.5 2005/03/25 11:05:00 mr Exp $

        Copyright (C) 2005 Marcus Rickert

        http://www.the-rickerts.de/mr/PAMINA.html
*/

/* ***INCLUDES************************************************************************ */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <TArray.h>
#include <TStatistics.h>

/* ***DEFINES************************************************************************* */

#define ParentClass TBaseObject

/* ***METHODS************************************************************************* */

TStatistics::TStatistics(long theMaxNrOfElements, int theNrOfCols)

{
    int i;
    
    NrOfCols = theNrOfCols;
    MaxNrOfElements = theMaxNrOfElements;
    if (MaxNrOfElements != -1)
        Data.Resize(NrOfCols * MaxNrOfElements);
    InputPointer = DeletePointer = NrOfElements = 0;
    Sums = (double *) malloc(NrOfCols * sizeof(double));
    Minima = (double *) malloc(NrOfCols * sizeof(double));
    Maxima = (double *) malloc(NrOfCols * sizeof(double));
    Deltas = (double *) malloc(NrOfCols * sizeof(double));
    Transients = (double *) malloc(NrOfCols * sizeof(double));
    Delays = (int *) malloc(NrOfCols * sizeof(int));
    DelayInits = (int *) malloc(NrOfCols * sizeof(int));
    for (i = 0; i < NrOfCols; i++)
    {
        Sums[i] = 0.0;
        Delays[i] = 0;
        Transients[i] = 0.0;
        Deltas[i] = 0.0;
        Minima[i] = 0.0;
        Maxima[i] = 0.0;
        DelayInits[i] = 1;
    }
}

/* =================================================================================== */

TStatistics::~TStatistics(void)

{
    Free(Sums);
    Free(Minima);
    Free(Maxima);
    Free(Deltas);
    Free(Transients);
    Free(Delays);
    Free(DelayInits);
}

/* =================================================================================== */

bool
TStatistics::AddValue(double Value, int Col)

{
    if (MaxNrOfElements == -1)
    {
        if (!Data.Add(Value))
            return FALSE;
    }
    else
    {
        if (NrOfElements == MaxNrOfElements)
            RemoveElement();
        else
            NrOfElements++;
        Data[InputPointer] = Value;
        WrapAround(InputPointer, MaxNrOfElements);
    }
    Sums[Col] += Value;
    if (NrOfElements == 1)
    {
        Minima[Col] = Value;
        Maxima[Col] = Value;
        Transients[Col] = Minima[Col];
        Delays[Col] = 0;
    }
    else
    {
        if (Value < Minima[Col])
        {
            Minima[Col] = Value;
            UpdateTransient(Col);
        }
        if (Value > Maxima[Col])
            Maxima[Col] = Value;
    }
    return TRUE;
}

/* =================================================================================== */

TString
TStatistics::GetClassName(void) const

{
    return "TStatistics";
}

/* =================================================================================== */

TString
TStatistics::GetInstanceName(void) const

{
    char line[128];

    sprintf(line, "#Cols:%d, #Elements:%ld, #MaxElements:%ld", NrOfCols, NrOfElements,
            MaxNrOfElements);
    return line;
}

/* =================================================================================== */

double
TStatistics::GetMinimum(int Col)

{
    return Minima[Col];
}

/* =================================================================================== */

double
TStatistics::GetMaximum(int Col)

{
    return Maxima[Col];
}

/* =================================================================================== */

double
TStatistics::GetMean(int Col)

{
    if (NrOfElements)
        return (double) Sums[Col] / (double) NrOfElements;
    else
        return -1.0;
}

/* =================================================================================== */

double
TStatistics::GetTransientMinimum(int Col)

{
    if (Delays[Col])
    {
        Delays[Col]--;
        Transients[Col] += Deltas[Col];
    }
    return Transients[Col];
}

/* =================================================================================== */

void
TStatistics::RemoveElement(int Col)

{
    long i;
    
    Sums[Col] -= Data[DeletePointer];
    if (EPSILON(Data[DeletePointer] - Minima[Col]))
    {
        Minima[Col] = 1e99;
        Data[DeletePointer] = 1e99;
        for (i = 0; i < NrOfElements; i++)
            if (Data[DeletePointer] < Minima[Col])
                Minima[Col] = Data[DeletePointer];
        UpdateTransient(Col);
    }
    if (EPSILON(Data[DeletePointer] - Maxima[Col]))
    {
        Maxima[Col] = -1e99;
        Data[DeletePointer] = -1e99;
        for (i = 0; i < NrOfElements; i++)
            if (Data[DeletePointer] > Maxima[Col])
                Maxima[Col] = Data[DeletePointer];
    }
    WrapAround(DeletePointer, MaxNrOfElements);
}

/* =================================================================================== */

void
TStatistics::SetDelay(int Delay, int Col)

{
    DelayInits[Col] = Delay;
}

/* =================================================================================== */

void
TStatistics::UpdateTransient(int Col)

{
    Delays[Col] = DelayInits[Col];
    Deltas[Col] = (Minima[Col] - Transients[Col]) / (double) Delays[Col];
}

/* =================================================================================== */
#undef ParentClass
#define ParentClass TStatistics
/* =================================================================================== */

TRealTimeStatistics::TRealTimeStatistics(long theMaxNrOfElements, double thePeriod,
                                         int theNrOfCols) :

ParentClass(theMaxNrOfElements, theNrOfCols)

{
    Interval = thePeriod / (double) theMaxNrOfElements;
    Next = 0.0;
    Timer.Start();
}

/* =================================================================================== */

bool
TRealTimeStatistics::AddValue(double Value, int Col)

{
    if (Timer.GetElapsedTime() > Next)
    {
        Next += Interval;
        return ParentClass::AddValue(Value, Col);
    }
    else
        return TRUE;
}

/* =================================================================================== */
#undef ParentClass
