/*      file     : parcontrol/TApplicationMaster.C
        project  : Parallel Toolbox 2.x
        author   : Marcus Rickert
        RCS      : $Id: TApplicationMaster.C,v 1.10 2005/03/25 11:07:01 mr Exp $

        Copyright (C) 2005 Marcus Rickert

        http://www.the-rickerts.de/mr/PAMINA.html
*/

// ***INCLUDES************************************************************************ 

#include <ParToolsConfig.h>
#include <ParControlConfig.h>

#include <TCommunication.h>
#include <TCPNInfo.h>
#include <TTopology.h>

#include <TClassInfo.h>
#include <TAdjacenceInfo.h>
#include <TObjectManager.h>
#include <TStaticBalancer.h>
#include "TApplicationSlave.h"
#include "TApplicationMaster.h"

// ***METHODS************************************************************************* 

// =================================================================================== 
#define ParentClass TBaseObject
// =================================================================================== 

TApplicationMaster::TApplicationMaster(TApplicationSlave *theApplicationSlave)

{
    ApplicationSlave = theApplicationSlave;
    Topology = 0;
    debug(Print(DM_LIFESPAN, "CONSTRUCTOR"));
}

// =================================================================================== 

TApplicationMaster::~TApplicationMaster(void)

{
    debug(Print(DM_LIFESPAN, "DESTRUCTOR"));
}

// =================================================================================== 

bool
TApplicationMaster::Activate(void)

{
    debug(Print(DM_LIFESPAN, "Activate"));
    Communication->RegisterRecipient(MTAG_APP_MASTER, this);
    Parallel->ShowHosts();
    if (!Parallel->AddAllCPN())
    {
        Parallel->Quit();
        return FALSE;
    }
    Parallel->ShowCPN();
    return TRUE;
}

// =================================================================================== 

bool
TApplicationMaster::HandleEvent(TEventID EventID, TBaseMessage *, int &Handled)

{
    switch(EventID)
    {
    case MEVENT_APP_MASTER_ACK_SEQUENCE:
        return HandleEventAckSequence(Handled);

    default:
        Error("illegal event ID %ld", EventID);
        return FALSE;
    }
#ifndef COMPLETE_SWITCH_FLOW
    return TRUE;
#endif
}

// =================================================================================== 

bool
TApplicationMaster::HandleEventAckSequence(int &Handled)

{
    MissingSeqAcks--;
    if (MissingSeqAcks < 0)
    {
        Error("HandleEventSeqAck: received superfluous sequence ack");
        return FALSE;
    }
    if (MissingSeqAcks == 0)
        Handled = EXIT_LOOP;
    debug(Print(DM_TIME_STEP, "HandleEventAckSeq: still missing: %ld", MissingSeqAcks));
    return TRUE;
}

// =================================================================================== 

bool
TApplicationMaster::DistributeSimulationObjects(TTopology  *theTopology,
                                                TNodeList  *Nodes, 
                                                TClassInfo *LocationClassInfo,
                                                TEdgeList  *Edges, 
                                                TClassInfo *AdjacenceClassInfo,
                                                TInfoMask   AdditionalNodeLoadMask,
                                                TInfoMask   AdditionalEdgeLoadMask)

{
    TStaticBalancer *Balancer = 0;
    int              CPNIndex;
    TNodeList       *NodesOnCPN;
    TEdgeList       *EdgesOnCPN;
    TInfoMask        NodeLoadMask;
    TInfoMask        EdgeLoadMask;

    Topology = theTopology;
    Balancer = new TStaticBalancer(Nodes, LocationClassInfo->GetInfoID(),
                                   Edges, AdjacenceClassInfo->GetInfoID(), 
                                   Topology, FALSE);
    NodeLoadMask = LocationClassInfo->GetInfoMask() | AdditionalNodeLoadMask;
    EdgeLoadMask = AdjacenceClassInfo->GetInfoMask() | AdditionalEdgeLoadMask;
    if (!(ObjectManager->FillContainers((TContainerObjectList *) Nodes, NodeLoadMask) AND
          ObjectManager->FillContainers((TContainerObjectList *) Edges, EdgeLoadMask)))
    {
        Error("cannot fill nodes and edges");
        return FALSE;
    }
    if (!Balancer->OrthogonalBisection())
    {
        Error("DistributeSimulationObjects: cannot bisect");
        return FALSE;
    }    
    for (CPNIndex = 0; CPNIndex < Topology->GetNrOfCPN(); CPNIndex++)
    {
        NodesOnCPN = Balancer->GetNodesOnCPN(CPNIndex);
        EdgesOnCPN = Balancer->GetEdgesOnCPN(CPNIndex);
        Print(DM_CONTROL, "assigning %ld nodes and %ld edges to CPN %s",
              NodesOnCPN->GetNrOfElements(), EdgesOnCPN->GetNrOfElements(), 
              (char *) Topology->GetCPNInfoByRank(CPNIndex)->GetInstanceName());
        if (!ObjectManager->LoadDataLocally())
        {
            Print(DM_CONTROL, "fill %ld nodes", NodesOnCPN->GetNrOfElements());
            if (!ObjectManager->FillContainers((TContainerObjectList *) NodesOnCPN))
            {
                Error("cannot fill nodes");
                return FALSE;
            }
            Print(DM_CONTROL, "fill %ld edges", EdgesOnCPN->GetNrOfElements());
            if (!ObjectManager->FillContainers((TContainerObjectList *) EdgesOnCPN))
            {
                Error("cannot fill edges");
                return FALSE;
            }
        }
        if (!ApplicationSlave->MigrateContainerObjects(Topology->GetCPNInfoByRank(CPNIndex),
                                                       (TContainerObjectList *) NodesOnCPN))
        {
            Error("cannot transfer nodes and edges to CPN %d", CPNIndex);
            return FALSE;
        }
    }
    DeleteObject(Balancer);
    return ApplicationSlave->CleanUpAfterTransfer();
}

// =================================================================================== 

bool
TApplicationMaster::SendEventToTopology(TEventID EventID, const char *Description,
                                        bool Synchronize)

{
    if (Topology == 0)
    {
        Error("SendEventToSlaves: no topology set");
        return FALSE;
    }
    Print(DM_CONTROL, "%s on %ld CPN...", Description, Topology->GetNrOfCPN());
    if (!Communication->SendEvent(MEVENT_APP_SLAVE_INITIALIZE_DATA, Topology->GetList(),
                                   MTAG_APP_SLAVE, MTYPE_APP_SLAVE))
    {
        Error("SendEventToTopology: cannot send event %d", EventID);
        return FALSE;
    }
    if (Synchronize)
    {
        if (!Parallel->Synchronize(Topology->GetList()))
        {
            Error("SendEventToTopology: cannot synchronize");
            return FALSE;
        }
        Print(DM_CONTROL, "Synchronized.");
    }
    return TRUE;
}

// =================================================================================== 

bool
TApplicationMaster::StartApplicationSequence(TTimeStep &Start, long NrOfTimeSteps)

{
    TBaseMessage *Message;
    TDuration     Duration(Start, NrOfTimeSteps);

    if (!Communication->GetNewMessage(Message, Topology->GetList(),
                                      MTAG_APP_SLAVE, MTYPE_APP_SLAVE))
    {
        Error("StartApplicationSequence: cannot get new ");
        return FALSE;
    }
    if (!(Message->Event(MEVENT_APP_SLAVE_START_SEQUENCE) AND
          Duration.Encode(Message) AND
          Communication->Commit(Message)))
    {
        Error("StartApplicationSequence: cannot send message");
        return FALSE;
    }
    MissingSeqAcks = Topology->GetNrOfCPN();
    if (!Communication->MessageLoop(TRUE))
    {
        Error("StartApplicationSequence: while waiting for end of sequence");
        return FALSE;
    }
    Start = Start + NrOfTimeSteps;
    return TRUE;
}

// =================================================================================== 

bool
TApplicationMaster::Run(void)

{
    if (!ApplicationSlave->Run())
        return FALSE;
    return TRUE;
}

// =================================================================================== 
#undef ParentClass

