/*      file     : parcontrol/TApplicationSlave.C
        project  : Parallel Toolbox 2.x
        author   : Marcus Rickert
        RCS      : $Id: TApplicationSlave.C,v 1.18 2005/03/25 11:07:02 mr Exp $

        Copyright (C) 2005 Marcus Rickert

        http://www.the-rickerts.de/mr/PAMINA.html
*/

// ***INCLUDES************************************************************************ 

#include <ParToolsConfig.h>
#include <ParControlConfig.h>
#include <ParStatTypes.h>
#include <ParControlTypes.h>

#include <TTimer.h>
#include <TSynchronize.h>
#include <TEnvironment.h>
#include <TExtendedCPNInfo.h>
#include <TParallel.h>
#include <TConfiguration.h>
#include <TCommunication.h>
#include <TParallel.h>
#include <TNetwork.h>
#include <TBaseDataManager.h>
#include <TObjectManager.h>
#include <TNode.h>
#include <TEdge.h>
#include <TIncidenceInfo.h>
#include <TStatManager.h>
#include <TDynamicBalancer.h>
#include <TStatManager.h>
#include <TSamplingDescription.h>
#include <TAggregationInfo.h>
#include <TAggregationResult.h>
#include <TAppEventHandler.h>
#include <TMapStat.h>

#include "TApplicationSlave.h"

#define PAR_CONTROL_DEFAULT_MAPSTATINTERVAL 0L
#define PAR_CONTROL_DEFAULT_SYNCPIPENAME    "/tmp/ParControlSync"

// ***METHODS************************************************************************* 

// =================================================================================== 
#define ParentClass TBaseObject
// =================================================================================== 

TApplicationSlave::TApplicationSlave(int argc, char *argv[])

{
    Initialize();
    AppEventHandler = 0;
    AppSync = 0;
    MapStatSamplingDescription = 0;
    MapStatAggregationInfo = 0;
    MyEnvironment = new TEnvironment(argc, argv);
    MyCommunication = new TCommunication;
    MyParallel = new TParallel;
    MyNetwork = new TNetwork;
    MyObjectManager = new TObjectManager;
    MyBaseDataManager = new TBaseDataManager;
    MyStatManager = new TStatManager;
    ApplicationTimer = new TTimer(Environment->GetOverallTimer());
    StatTimer = new TTimer(ApplicationTimer);
    NodeTimer = new TTimer(ApplicationTimer);
    EdgeTimer = new TTimer(ApplicationTimer);
    DynamicBalancer = new TDynamicBalancer(&ApplicationNodes, &ApplicationEdges, 
                                           &BoundaryEdges, NodeTimer, EdgeTimer,
                                           StatTimer);
    debug(Print(DM_LIFESPAN, "CONSTRUCTOR"));    
}

// =================================================================================== 

TApplicationSlave::~TApplicationSlave(void)

{
    debug(Print(DM_LIFESPAN, "DESTRUCTOR"));
    Print(DM_TIME, "nodes:%.1f [s], edges:%.1f [s], stat:%.1f [s], all:%.1f [s]",
          NodeTimer->GetElapsedTime(), EdgeTimer->GetElapsedTime(),
          StatTimer->GetElapsedTime(), ApplicationTimer->GetElapsedTime());
    ApplicationNodes.DeleteAll(TRUE);
    ApplicationEdges.DeleteAll(TRUE);
    DeleteObject(EdgeTimer);
    DeleteObject(NodeTimer);
    DeleteObject(StatTimer);
    DeleteObject(ApplicationTimer);
    DeleteObject(DynamicBalancer);
    DeleteObject(MyStatManager);
    DeleteObject(MyBaseDataManager);
    DeleteObject(MyObjectManager);
    DeleteObject(MyNetwork);
    DeleteObject(MyParallel);
    DeleteObject(MyCommunication);
    DeleteObject(MyEnvironment);
    DeleteObject(MapStatSamplingDescription);
    DeleteObject(MapStatAggregationInfo);
    DeleteObject(AppEventHandler);
    DeleteObject(AppSync);
}

// =================================================================================== 

bool
TApplicationSlave::Activate(void)

{
    debug(Print(DM_LIFESPAN, "ACTIVATE"));
    if (!(Environment->Activate() AND
          Communication->Activate() AND
          Parallel->Activate() AND
          ObjectManager->Activate() AND
          MyBaseDataManager->Activate() AND
          StatManager->Activate() AND
          DynamicBalancer->Activate()))
        return FALSE;
    TNode::ClassInfo->RegisterActionHandler(this);
    TEdge::ClassInfo->RegisterActionHandler(this);    
    Communication->RegisterRecipient(MTAG_APP_SLAVE, this);
    ObjectManager->RegisterClass(0, TMapStat::ClassInfo,
                                 CLASS_EXTENDED_CPNINFO, CLASS_MAP_STAT, this);
    Configuration->SetDefault(TConfigEntry(PAR_CONTROL_GROUP,
                                           PAR_CONTROL_NAME_MAPSTATINTERVAL,
                                           PAR_CONTROL_DEFAULT_MAPSTATINTERVAL));
    return TRUE;
}

// =================================================================================== 

bool
TApplicationSlave::ActivateAppEventHandler(void)

{
    if (Configuration->Exists(PAR_CONTROL_GROUP, PAR_CONTROL_NAME_EVENTFILENAME))
    {
        TString Filename;

        Filename = Environment->ReplacePathTokens(Configuration->GetString());
        AppEventHandler = new TAppEventHandler;
        if (!AppEventHandler->Open(Filename))
        {
            Error("ActivateAppEventHandler: cannot activate handler");
            return FALSE;
        }
    }
    return TRUE;
}

// =================================================================================== 

bool
TApplicationSlave::ActivateStatistics(void)

{    
    MapStatInterval = (int) Configuration->GetLong(PAR_CONTROL_GROUP,
                                                   PAR_CONTROL_NAME_MAPSTATINTERVAL);
    if (MapStatInterval > 0)
    {
        MapStatSamplingDescription = 
             new TSamplingDescription(TMapStat::ClassInfo->GetID(), 0,
                                      DEST_ALL, SAMPLING_ID_MAP_STAT);
        MapStatAggregationInfo = new TAggregationInfo(this, MapStatSamplingDescription);
        MapStatSamplingDescription->SetCollectInterval(MapStatInterval);
        StatManager->RegisterAggregation(MapStatAggregationInfo);
        StatManager->StartSampling(MapStatSamplingDescription);
    }   
    if (!StatManager->RegisterSamplingHandler(TMapStat::ClassInfo->GetID(), this))
    {
        Error("Run: cannot register myself as sampling handler");
        return FALSE;
    }
    return TRUE;
}

// =================================================================================== 

bool
TApplicationSlave::CleanUpAfterTransfer(void)

{
    if (!ObjectManager->Migrate())
    {
        Error("in CleanUpAfterTransfer");
        return FALSE;
    }
    if (!Communication->SendEvent(MEVENT_APP_SLAVE_TRANSFER_COMPLETE,
                                  &TransferDestinations, MTAG_APP_SLAVE,
                                  MTYPE_APP_SLAVE))
    {
        Error("CleanUpAfterTransfer: cannot send MEVENT_APP_SLAVE_TRANSFER_COMPLETE");
        return FALSE;
    }
    if (!Parallel->Synchronize(&TransferDestinations))
    {
        Error("CleanUpAfterTransfer: cannot synchronize");
        return FALSE;
    }
    TransferDestinations.DeleteAll();
    return TRUE;
}

// =================================================================================== 

bool
TApplicationSlave::Collect(TSamplingDescription *Description)

{
    if (Description->GetStatObjectClassID() == TMapStat::ClassInfo->GetID())
    {
        TMapStat *Stat = new TMapStat(Parallel->GetLocalCPNID());

        Stat->SetTimeStep(Description->GetTimeStep());
        if (!Stat->Collect(&ApplicationNodes, &ApplicationEdges))
        {
            Error("Collect: cannot collect map stat");
            return FALSE;
        }
        Description->GetObjectList()->Append(Stat);
        return TRUE;
    }
    return TStatHandler::Collect(Description);
}

// =================================================================================== 

bool
TApplicationSlave::ExecuteSubTimeStep(void)

{
    debug(Print(DM_TIME_STEP, "ExecuteTimeStep: %s",
                (char *) ApplicationTimeStep.GetInstanceName()));
    StatTimer->Select();
    if (!StatManager->ExecuteSubTimeStep(ApplicationTimeStep))
    {
        Error("ExecuteSubTimeStep: while calling statistics manager");
        return FALSE;
    }
    StatTimer->Unselect();
    if (!DynamicBalancer->ExecuteSubTimeStep(ApplicationTimeStep))
    {
        Error("ExecuteSubTimeStep: while calling load balancer");
        return FALSE;
    }
    if (!InitializeDataAfterReception())
    {
        Error("ExecuteSubTimeStep: while initializing data");
        return FALSE;
    }
    InitNodes.DeleteAll();
    InitEdges.DeleteAll();
    if (ApplicationTimeStep.GetSubStep() == 0)
    {
        if (AppEventHandler)
            if (!AppEventHandler->ExecuteTimeStep(ApplicationTimeStep))
            {
                Error("while handling application event");
                return FALSE;
            }
        if (AppSync)
        {
            long WaitStep;
            bool ErrorFlag;
            
            if (!AppSync->Step(ApplicationTimeStep))
            {
                Error("ExecuteSubTimeStep: cannot step");
                return FALSE;
            }
            WaitStep = ApplicationTimeStep;
            if (MasterApplication)
                WaitStep--;
            while (AppSync->Wait(WaitStep, ErrorFlag))
            {
                IdleTime();
            }       
            if (ErrorFlag)
            {
                Error("ExecuteSubTimeStep: during wait for step %ld", WaitStep);
                return FALSE;
            }
        }
    }
    return TRUE;
}

// =================================================================================== 

TString
TApplicationSlave::GetClassName(void) const

{
    return "TApplicationSlave";
}

// =================================================================================== 

TString
TApplicationSlave::GetInstanceName(void) const

{
    char Line[256];

    sprintf(Line, "N:%ld E:%ld BE:%ld", ApplicationNodes.GetNrOfElements(),
            ApplicationEdges.GetNrOfElements(), BoundaryEdges.GetNrOfElements());
    return Line;
}

// =================================================================================== 

bool
TApplicationSlave::HandleAction(TParallelObject *Object, TObjectAction Action)

{
    if (Object->GetClassID() == TNode::ClassInfo->GetID())
    {

        TNode *Node = (TNode *) Object;

        switch (Action)
        {
        case ActionMerge:

            Warning("MERGING container!");

        case ActionAdd:
            
            ApplicationNodes.Append(Node);
            InitNodes.Append(Node);
            return TRUE;

        case ActionRemove:

            ApplicationNodes.Delete(Node);
            InitNodes.Delete(Node);
            return TRUE;

        default:
            //      Warning("action %d ignored", (int) Action);
            ;
        }
    } 
    else if (Object->GetClassID() == TEdge::ClassInfo->GetID())
    {
        TEdge *Edge = (TEdge *) Object;

        switch (Action)
        {
        case ActionAdd:     
            
            NonBoundaryEdges.Append(Edge);
            ApplicationEdges.Append(Edge);
            
        case ActionMerge:
        case ActionSplit:
            
            InitEdges.Append(Edge);
            return TRUE;
            
        case ActionRemove:

            InitEdges.Delete(Edge);
            NonBoundaryEdges.Delete(Edge);
            ApplicationEdges.Delete(Edge);
            if (BoundaryEdges.IsInList(Edge))
                Warning("******** EDGE %s WAS STILL in BOUNDARY LIST ********",
                        (char *) Edge->GetInstanceName());
            BoundaryEdges.Delete(Edge);
            return TRUE;

        case ActionAddBoundary:

            NonBoundaryEdges.Delete(Edge);
            BoundaryEdges.Append(Edge);
            return TRUE;

        case ActionRemoveBoundary:

            NonBoundaryEdges.Append(Edge);
            BoundaryEdges.Delete(Edge);
            return TRUE;

        default:
            //      Warning("action %d ignored", (int) Action);
            ;
        } 
    }
    return TRUE;
}

// =================================================================================== 

bool
TApplicationSlave::HandleAggregationResult(TAggregationInfo *AggregationInfo, 
                                           TTimeStep &TimeStep)

{
    if (AggregationInfo == MapStatAggregationInfo)
        return HandleMapStatResult(MapStatAggregationInfo, TimeStep);
    return TStatHandler::HandleAggregationResult(AggregationInfo, TimeStep);
}

// =================================================================================== 

bool
TApplicationSlave::HandleCreateRequest(TParallelObject *&Object, const TClassInfo *ClassInfo, 
                                       TParallelID ParallelID)

{
    if (ClassInfo == TMapStat::ClassInfo)

        Object = new TMapStat(ParallelID);

    return TRUE;
}

// =================================================================================== 

bool
TApplicationSlave::HandleEvent(TEventID EventID, TBaseMessage *Message, int &Handled)

{
    switch(EventID)
    {

    case MEVENT_APP_SLAVE_INITIALIZE_DATA:
        Handled = TRUE;
        return FALSE;

    case MEVENT_APP_SLAVE_START_SEQUENCE:
        Handled = TRUE;
        return HandleEventStartApplicationSequence(Message);

    case MEVENT_APP_SLAVE_TRANSFER_COMPLETE:
        Handled = TRUE;
        return HandleEventInitializeDataAfterReception();

    default:
        Error("illegal event ID %ld", EventID);
        return FALSE;
    }
#ifndef COMPLETE_SWITCH_FLOW
    return TRUE;
#endif
}


// =================================================================================== 

bool
TApplicationSlave::HandleEventInitializeDataAfterReception(void)

{
    if (!InitializeDataAfterReception())
        return FALSE;
    InitNodes.DeleteAll();
    InitEdges.DeleteAll();
    return TRUE;
}

// =================================================================================== 

bool
TApplicationSlave::HandleEventStartApplicationSequence(TBaseMessage *Message)

{
    if (!ApplicationSequence.Decode(Message))
    {
        Error("HandleEventStartAppSeq: cannot decode duration");
        return FALSE;
    }
    Message->Deselect(); // poor PVM!
    ApplicationTimeStep = ApplicationSequence.GetStart();
    while (ApplicationTimeStep < ApplicationSequence.GetEnd())
    {
        if (!ExecuteSubTimeStep())
        {
            Error("HandleEventStartAppSeq: during time-step %s",
                  (char *) ApplicationTimeStep.GetInstanceName());
            return FALSE;
        }
        if (Parallel->GetCurrentNrOfCPN() == 1)
            if (!Communication->ProcessPendingMessages())
            {
                Error("ExecuteSubTimeStep: while in message loop (1 CPN)");
                return FALSE;
            }
        ApplicationTimeStep++;
    }
    return Communication->SendEvent(MEVENT_APP_MASTER_ACK_SEQUENCE, Message->GetSender(),
                                    MTAG_APP_MASTER, MTYPE_APP_MASTER);
}

// =================================================================================== 

bool
TApplicationSlave::HandleMapStatResult(TAggregationInfo *AggregationInfo, 
                                       TTimeStep &TimeStep)

{
    TAggregationResult *Result = 0;
    TMapStatList       *StatList = 0;
    TMapStatEnum        CStat;
    TInfoID             InfoID;

    if ((Result = (TAggregationResult *) AggregationInfo->GetResult(TimeStep)) == 0)
    {
        Error("HandleMapStatResult: cannot retrieve aggregation result");
        return FALSE;
    }
    if ((StatList = (TMapStatList *) Result->StatObjectList) == 0)
    {
        Error("HandleMapStatResult: cannot retrieve stat list");
        return FALSE;
    }
    InfoID = TMapStat::ClassInfo->GetInfoID();
    CStat.WalksForwardThrough(StatList);
    while (CStat())
    {
        TExtendedCPNInfo *Info;

        //      Print("subnetwork on CPN %ld has extent %s", CStat->GetParallelID(),
        //            (char *) CStat->GetMapRectangle().GetInstanceName());

        if ((Info = Network->GetExtendedCPNInfo(CStat->GetParallelID(), TRUE)) == 0)
        {
            Warning("HandleMapStatResult: cannot retrieve CPNInfo for %ld",
                    CStat->GetParallelID());
            return FALSE;
        }
        Info->SetInfo(InfoID, (TMapStat *) CStat);
        CStat++;
    }
    AggregationInfo->DisposeResult(TimeStep);
    if (!ObjectManager->Action(TMapStat::ClassInfo, 0, ActionModified))
    {
        Error("HandleMapStatResult: error while calling action handler");
        return FALSE;
    }
    return TRUE;
}

// =================================================================================== 

bool
TApplicationSlave::IdleTime(void)

{
    return TRUE;
}

// =================================================================================== 

bool
TApplicationSlave::InitializeDataAfterReception(void)

{
    TEdgeEnum CEdge;
    TNodeEnum CNode;

    /*
    if (InitNodes.GetNrOfElements() + InitEdges.GetNrOfElements() > 0)
    {
        Print(DM_CONTROL, "initialize %ld nodes and %ld edges",
              InitNodes.GetNrOfElements(), InitEdges.GetNrOfElements());
        CEdge.WalksForwardThrough(InitEdges);
        while (CEdge())
        {
            if (!MyBaseDataManager->InsertIncidences(CEdge))
                return FALSE;
            CEdge++;
        }
        CNode.WalksForwardThrough(InitNodes);
        while (CNode())
        {
            if (!MyBaseDataManager->GenerateIncidenceInfo(CNode))
            return FALSE;
            CNode++;
        }
    }
    */
    return TRUE;
}

// =================================================================================== 

bool
TApplicationSlave::Run(void)

{
    return (Environment->Run() AND
            ObjectManager->Run() AND
            MyBaseDataManager->Run() AND
            Parallel->Start() AND
            StatManager->Run(Parallel->GetPhysicalTopology()) AND
            DynamicBalancer->Run() AND
            SynchronizeApplications() AND
            ActivateStatistics() AND
            ActivateAppEventHandler());
}

// =================================================================================== 

bool
TApplicationSlave::MigrateContainerObjects(const TCPNInfo *Destination, 
                                           TContainerObjectList *List)

{
    TransferDestinations.Append(Destination);
    return ObjectManager->MigrateContainerObjects(Destination, List);
}

// =================================================================================== 

bool
TApplicationSlave::SynchronizeApplications(void)

{
    Configuration->SetDefault(TConfigEntry(PAR_CONTROL_GROUP, PAR_CONTROL_NAME_SYNCPIPENAME,
                                           PAR_CONTROL_DEFAULT_SYNCPIPENAME));
    SlaveApplication = Configuration->OptionActive("s");
    MasterApplication = Configuration->OptionActive("m");
    if (Parallel->IsMaster() AND (SlaveApplication OR MasterApplication))
    {   
        AppSync = new TSynchronize;
        Print("waiting for synchronization...");
        if (!AppSync->Activate(Configuration->GetString(PAR_CONTROL_GROUP, 
                                                        PAR_CONTROL_NAME_SYNCPIPENAME),
                               MasterApplication))
        {
            Error("SynchronizeApplications: cannot activate synchronization");
            return FALSE;
        }
        Print("synchronization established");
    }
    return TRUE;
}

// =================================================================================== 
#undef ParentClass


